# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
A simple parser library based on recursive descent and operator precedence parsing.
Recursive decent is performed by looking at the next token,
and then dispatching through a table of token kinds to select the appropriate Rule.

Upon parsing, each type of rule produces a generic result, which is then passed to a user-provided transformer function.
Different rule types have different transformer function types to match the shape of the parsed structure.
Transformers can return results of any type.

Several types of rules are available:
* Atom: a single token kind.
* Prefix: a prefix token, followed by a body rule, optionally followed by a suffix.
* Quantity: a body rule, optionally interleaved with a separator token, with optional `min` and `max` quantities.
* Choice: one of several rules, which must be distinguishable by the head token.
* Precedence: a family of binary precedence operators.

The main design feature of Parser is that operators are not first-class rules:
they are parsed within the context of a Precedence rule.
This allows us to use the simple operator-precedence parsing algorithm,
while expressing other aspects of a grammar using straightforward recursive descent.
'''

from collections import namedtuple
from typing import (Any, Callable, Dict, FrozenSet, Iterable, Iterator, List, NoReturn, Optional, Tuple, Type, TypeVar, Union,
  cast)

from tolkien import Source, Syntax, SyntaxMsg, Token

from .buffer import Buffer
from .graph import visit_nodes
from .io import tee_to_err
from .lex import Lexer, reserved_names, valid_name_re
from .meta import get_caller_module_name
from .string import indent_lines, iter_str, pluralize
from .untyped import Immutable


TokenKind = str
RuleName = str
RuleRef = Union['Rule',RuleName]
_T = TypeVar('_T')


class ParseError(Exception):
  error_prefix = 'parse'

  def __init__(self, source:Source, syntax:Syntax, msg:str, notes:Iterable[SyntaxMsg]=()) -> None:
    self.source = source
    self.notes:List[SyntaxMsg] = list(notes)
    self.syntax = syntax
    self.msg = msg
    super().__init__((self.syntax, self.msg))

  def fail(self) -> NoReturn:
    self.source.fail(*reversed(self.notes), (self.syntax, f'{self.error_prefix} error: {self.msg}'))

  def add_in_note(self, syntax:Syntax, context:Any) -> None:
    # For repeated positions, keep only the innermost note.
    # Outer notes are for choice, quantity, and optional, which are not very informative for debugging parse errors.
    if not self.notes or self.notes[-1][0] != syntax:
      self.notes.append((syntax, f'note: in {context}:'))


class ExcessToken(ParseError):
  'Raised by Parser when expression parsing completes but does not exhaust the token stream.'


def append(list_:List[_T], el:_T) -> List[_T]:
  'Append an element to a list and return the list. Useful when writing transform lambdas.'
  list_.append(el)
  return list_

def append_or_list(list_or_el:Union[_T,List[_T]], el:_T) -> List[_T]:
  'Create a list from two elements, or if the left element is already a list, append the right element to it and return it.'
  if isinstance(list_or_el, list):
    list_or_el.append(el)
    return list_or_el
  else:
    return [list_or_el, el]


TokenTransform = Callable[[Source,Token],Any]

def token_identity(source:Source, token:Token) -> Token: return token

def token_extract_text(source:Source, token:Token) -> str: return source[token]
def token_extract_kind(source:Source, token:Token) -> str: return token.kind


UnaryTransform = Callable[[Source,Token,Any],Any]
def unary_syn(source:Source, token:Token, obj:Any) -> Tuple[str,Any]: return (source[token], obj)
def unary_identity(source:Source, token:Token, obj:Any) -> Any: return obj

BinaryTransform = Callable[[Source,Token,Any,Any],Any]
def binary_syn(source:Source, token:Token, left:Any, right:Any) -> Tuple[str,Any,Any]: return (source[token], left, right)
def adjacency_syn(source:Source, token:Token, left:Any, right:Any) -> Tuple[Any,Any]: return (left, right)
def binary_to_list(source:Source, token:Token, left:Any, right:Any) -> List[Any]: return append_or_list(left, right)

QuantityTransform = Callable[[Source,Token,List[Any]],Any]
def quantity_identity(source:Source, token:Token, elements:List[Any]) -> List[Any]: return elements

StructTransform = Callable[[Source,Token,List[Any]],Any]
def struct_syn(source:Source, token:Token, fields:List[Any]) -> Tuple[Any,...]: return tuple(fields)

ChoiceTransform = Callable[[Source,Token,RuleName,Any],Any]
def choice_syn(source:Source, token:Token, label:RuleName, obj:Any) -> Tuple[str,Any]: return (label, obj)
def choice_identity(source:Source, token:Token, label:RuleName, obj:Any) -> Any: return obj

_sentinel_kind = '!SENTINEL'



class Rule:
  'A parser rule. A complete parser is created from a graph of rules.'

  name:str # Optional name, used to link the rule graph.
  type_desc:str # Type description for diagnostic descriptions.
  sub_refs:Tuple[RuleRef,...] = () # The rules or references (name strings) that this rule refers to.
  subs:Tuple['Rule',...] = () # Sub-rules, obtained by linking sub_refs.
  heads:Tuple[TokenKind,...] # Set of leading token kinds for this rule.

  def __init__(self, *args:Any, **kwargs:Any) -> None: raise Exception(f'abstract base class: {self}')


  def __str__(self) -> str:
    if self.name: return f'{self.name!r} {self.type_desc}'
    else: return repr(self)


  def __repr__(self) -> str:
    parts = []
    if self.name: parts.append(f'name={self.name!r}')
    parts.extend((repr(s) for s in self.sub_refs))
    s = ', '.join(parts)
    return f'{type(self).__name__}({s})'


  def __lt__(self, other:'Rule') -> bool:
    if not isinstance(other, Rule): raise ValueError(other)
    return str(self) < str(other)

  @property
  def subs_desc(self) -> str:
    return f'[{", ".join(str(s) for s in self.subs)}]'


  def token_kinds(self) -> Iterable[str]:
    'The token kinds directly referenced by this rule.'
    return ()


  def head_subs(self) -> Iterable['Rule']:
    'The sub-rules that determine the head of a match for this rule. Used by `compile_heads`.'
    raise NotImplementedError(self)


  def compile_heads(self) -> Iterator[TokenKind]:
    'Calculate the `heads` set for this rule. The parser uses this to build a dispatch table against token kinds.'
    if self.heads: # For recursive rules, the sentinel causes recursion to break here.
      yield from self.heads
      return
    # Otherwise, this rule does not yet know its own heads; discover them recursively.
    self.heads = (_sentinel_kind,) # Temporarily set to prevent recursion loops; use an impossible name.
    for sub in self.head_subs():
      yield from sub.compile_heads()
    # Now reset self.heads; it is up to Parser to call `compile_heads` for each node.
    # Otherwise, we would be relying on possibly incomplete sets that accumulate in intermediates during deep recursion.
    self.heads = ()


  def compile(self, parser:'Parser') -> None: pass


  def expect(self, source:Source, token:Token, kind:str) -> Token:
    if token.kind != kind: raise ParseError(source, token, f'{self} expects {kind}; received {token.kind}.')
    return token


  def parse(self, parent:'Rule', source:Source, token:Token, buffer:Buffer[Token]) -> Any:
    raise NotImplementedError(self)


  def parse_sub(self, sub:'Rule', source:Source, start:Token, token:Token, buffer:Buffer[Token]) -> Any:
    try:
      return sub.parse(self, source, token, buffer)
    except ParseError as e:
      e.add_in_note(start, self)
      raise



class Atom(Rule):
  '''
  A rule that matches a single token.
  '''
  type_desc = 'atom'

  def __init__(self, kind:TokenKind, transform:TokenTransform=token_identity) -> None:
    self.name = ''
    self.heads = (kind,) # Pre-fill heads; compile_heads will return without calling head_subs, which Atom does not implement.
    self.kind = validate_name(kind)
    self.transform = transform


  def token_kinds(self) -> Iterable[str]:
    yield self.kind


  def parse(self, parent:Rule, source:Source, token:Token, buffer:Buffer[Token]) -> Any:
    parent.expect(source, token=token, kind=self.kind)
    #^ We use `parent` and not `self` to expect the token for a more contextualized error message.
    #^ Using `self` we get messages like "'newline' atom expects newline".
    return self.transform(source, token)



class _QuantityRule(Rule):
  'Base class for Opt and Quantity.'
  min:int
  body_heads:FrozenSet[str]
  drop:FrozenSet[str]

  @property
  def body(self) -> Rule:
    return self.subs[0]


  def head_subs(self) -> Iterable['Rule']:
    return (self.body,)


  def compile(self, parser:'Parser') -> None:
    self.body_heads = frozenset(self.body.heads)
    drop_body_intersect = self.body_heads & self.drop
    if drop_body_intersect:
      raise Parser.DefinitionError(f'{self} drop kinds and body head kinds intersect: {drop_body_intersect}')



class Opt(_QuantityRule):
  '''
  A rule that optionally matches another rule.
  '''
  type_desc = 'optional'
  min = 0

  def __init__(self, body:RuleRef, drop:Iterable[str]=(), dflt=None, transform:UnaryTransform=unary_identity) -> None:
    self.name = ''
    self.sub_refs = (body,)
    self.heads = ()
    self.body_heads = frozenset() # Replaced by compile.
    self.drop = frozenset(iter_str(drop))
    self.dflt = dflt
    self.transform = transform


  def parse(self, parent:Rule, source:Source, token:Token, buffer:Buffer[Token]) -> Any:
    res = self.dflt
    while token.kind in self.drop:
      token = next(buffer)
    if token.kind in self.body_heads:
      res = self.parse_sub(self.body, source, token, token, buffer)
    else:
      buffer.push(token)
    return self.transform(source, token, res)



class Quantity(_QuantityRule):
  '''
  A rule that matches some quantity of another rule.
  '''
  type_desc = 'quantity'

  def __init__(self, body:RuleRef, min:int, max:Optional[int], sep:TokenKind=None, sep_at_end:Optional[bool]=None, repeated_seps=False, drop:Iterable[str]=(),
   transform:QuantityTransform=quantity_identity) -> None:
    if min < 0: raise ValueError(min)
    if max is not None and max < 1: raise ValueError(max) # The rule must consume at least one token; see `parse` implementation.
    if sep is None and sep_at_end is not None: raise ValueError(f'`sep` is None but `sep_at_end` is `{sep_at_end}`')
    self.name = ''
    self.sub_refs = (body,)
    self.heads = ()
    self.sep = sep if sep is None else validate_name(sep)
    self.sep_at_end:Optional[bool] = sep_at_end
    self.repeated_seps = repeated_seps
    self.min = min
    self.max = max
    self.body_heads = frozenset() # Replaced by compile.
    self.drop = frozenset(iter_str(drop))
    self.transform = transform


  def token_kinds(self) -> Iterable[str]:
    if self.sep is not None:
      yield self.sep
    yield from self.drop


  def parse(self, parent:Rule, source:Source, token:Token, buffer:Buffer[Token]) -> Any:
    els:List[Any] = []
    found_sep = False
    start = token
    while True:
      while token.kind in self.drop:
        token = next(buffer)
      if len(els) == self.max or token.kind not in self.body_heads: break
      el = self.parse_sub(self.body, source, start, token, buffer)
      els.append(el)
      token = next(buffer)
      if self.sep is not None: # Parse separator.
        found_sep = (token.kind == self.sep)
        if found_sep:
          token = next(buffer)
        elif self.sep_at_end:
          raise ParseError(source, token, f'{self} expects {self.sep} separator; received {token.kind}.')
        else:
          break
        if self.repeated_seps:
          while token.kind == self.sep:
            token = next(buffer)

    if len(els) < self.min:
      body_plural = pluralize(self.min, f'{self.body} element')
      raise ParseError(source, token, f'{self} expects at least {body_plural}; received {token.kind}.')

    if self.sep_at_end is False and found_sep:
      raise ParseError(source, token, f'{self} received disallowed trailing {self.sep} separator.')

    buffer.push(token)
    return self.transform(source, start, els)



class ZeroOrMore(Quantity):
  def __init__(self, body:RuleRef, sep:TokenKind=None, sep_at_end:Optional[bool]=None, repeated_seps=False, drop:Iterable[str]=(),
   transform:QuantityTransform=quantity_identity) -> None:
    super().__init__(body=body, min=0, max=None, sep=sep, sep_at_end=sep_at_end, repeated_seps=repeated_seps, drop=drop, transform=transform)



class OneOrMore(Quantity):
  def __init__(self, body:RuleRef, sep:TokenKind=None, sep_at_end:Optional[bool]=None, repeated_seps=False, drop:Iterable[str]=(),
   transform:QuantityTransform=quantity_identity) -> None:
    super().__init__(body=body, min=1, max=None, sep=sep, sep_at_end=sep_at_end, repeated_seps=repeated_seps, drop=drop, transform=transform)



class Struct(Rule):
  '''
  A rule that matches a sequence of sub rules, producing a tuple of values.
  '''
  type_desc = 'structure'

  def __init__(self, *fields:RuleRef, drop:Iterable[str]=(), transform:StructTransform=None) -> None:
    if not fields: raise ValueError('Struct requires at least one field')
    self.name = ''
    self.sub_refs = fields
    self.heads = ()
    self.drop = frozenset(iter_str(drop))
    self.transform = transform


  def head_subs(self) -> Iterable['Rule']:
    for field in self.subs:
      yield field
      if not (isinstance(field, Quantity) and field.min == 0):
        break


  def compile(self, parser:'Parser') -> None:
    if self.transform is None:
      self.transform = parser._mk_struct_transform(name=self.name, subs=self.subs)


  def parse(self, parent:Rule, source:Source, token:Token, buffer:Buffer[Token]) -> Any:
    els:List[Any] = []
    start = token
    for i, field in enumerate(self.subs):
      if i:
        token = next(buffer)
      while token.kind in self.drop:
        token = next(buffer)
      el = self.parse_sub(field, source, start, token, buffer)
      els.append(el)
    assert self.transform is not None
    return self.transform(source, start, els)



class Choice(Rule):
  '''
  A rule that matches one of a set of choices, which must have unambiguous heads.
  '''
  type_desc = 'choice'

  def __init__(self, *choices:RuleRef, drop:Iterable[str]=(), transform:ChoiceTransform=choice_identity) -> None:
    self.name = ''
    self.sub_refs = choices
    self.heads = ()
    self.drop = frozenset(iter_str(drop))
    self.transform = transform
    self.head_table:Dict[TokenKind,Rule] = {}


  def head_subs(self) -> Iterable[Rule]: return self.subs


  def compile(self, parser:'Parser') -> None:
    for head in self.heads:
      matching_subs = [s for s in self.subs if head in s.heads]
      assert matching_subs
      if len(matching_subs) > 1:
        raise Parser.DefinitionError(f'{self} contains ambiguous choices for head token {head!r}:\n',
          *indent_lines(str(s) for s in matching_subs))
      self.head_table[head] = matching_subs[0]


  def parse(self, parent:Rule, source:Source, token:Token, buffer:Buffer[Token]) -> Any:
    start = token
    while token.kind in self.drop:
      token = next(buffer)
    try: sub = self.head_table[token.kind]
    except KeyError: pass
    else:
      syn = self.parse_sub(sub, source, start, token, buffer)
      return self.transform(source, start, sub.name, syn)
    exp = self.name or f'any of {self.subs_desc}'
    raise ParseError(source, token, f'{parent} expects {exp}; received {token.kind}.')



class Operator:
  'An operator that composes a part of a Precedence rule.'
  kinds:Tuple[TokenKind,...] = ()
  sub_refs:Tuple[RuleRef,...] = ()

  # TODO: spacing requirement options, e.g. no space, some space, symmetrical space.
  def __init__(self, *args:Any, **kwargs:Any) -> None: raise Exception(f'abstract base class: {self}')


  def parse_right(self, parent:Rule, source:Source, left:Any, op_token:Token, buffer:Buffer[Token], parse_level:Callable, level:int) -> Any:
    raise NotImplementedError(self)



class Suffix(Operator):
  'A suffix/postfix operator: the suffix follows the primary expression. E.g. `*` in `A*`.'

  def __init__(self, suffix:TokenKind, transform:UnaryTransform=unary_syn) -> None:
    self.kinds = (validate_name(suffix),)
    self.transform = transform


  def parse_right(self, parent:Rule, source:Source, left:Any, op_token:Token, buffer:Buffer[Token], parse_level:Callable, level:int) -> Any:
    return self.transform(source, op_token, left) # No right-hand side.



class SuffixRule(Operator):
  '''
  A suffix/postfix rule: like the Suffix operator, except the suffix is an arbitrary rule.
  Note: due to current limitations in the linker implementation,
  `suffix` must be a constructed rule and not a string reference.
  '''

  def __init__(self, suffix:Rule, transform:BinaryTransform=binary_syn) -> None:
    self.sub_refs = (suffix,)
    self.transform = transform


  @property
  def suffix(self) -> Rule: return cast(Rule, self.sub_refs[0]) # TODO: link over operators, and refer to self.subs instead.


  @property
  def kinds(self) -> Tuple[TokenKind,...]: # type: ignore
    return tuple(self.suffix.heads)


  def parse_right(self, parent:Rule, source:Source, left:Any, op_token:Token, buffer:Buffer[Token], parse_level:Callable, level:int) -> Any:
    right = parent.parse_sub(self.suffix, source, op_token, op_token, buffer)
    return self.transform(source, op_token.pos_token(), left, right)



class BinaryOp(Operator):
  'Abstract base class for binary operators that take left and right primary expressions.'



class Adjacency(BinaryOp):
  'A binary operator that joins two primary expressions with no operator token in between.'
  kinds:Tuple[TokenKind,...] = () # Adjacency operators have no operator token.

  def __init__(self, transform:BinaryTransform=adjacency_syn) -> None:
    self.transform = transform


  @property # type: ignore
  def kinds(self) -> Tuple[TokenKind,...]: # type: ignore
    raise _AllLeafKinds


  def parse_right(self, parent:Rule, source:Source, left:Any, op_token:Token, buffer:Buffer[Token], parse_level:Callable, level:int) -> Any:
    right = parse_level(parent=parent, source=source, token=op_token, buffer=buffer, level=level)
    return self.transform(source, op_token.pos_token(), left, right)



class _AllLeafKinds(Exception):
  'Raised by Adjacency.kinds to signal that the precedence parser associates the set of leaf token kinds with this adjacency op.'



class Infix(BinaryOp):
  'A binary operator that joins two primary expressions with an infix operator.'

  def __init__(self, kind:TokenKind, transform:BinaryTransform=binary_syn) -> None:
    self.kinds = (validate_name(kind),)
    self.transform = transform


  def parse_right(self, parent:Rule, source:Source, left:Any, op_token:Token, buffer:Buffer[Token], parse_level:Callable, level:int) -> Any:
    right = parse_level(parent=parent, source=source, token=next(buffer), buffer=buffer, level=level)
    return self.transform(source, op_token, left, right)



class Group:
  level_bump = 0
  'Operator precedence group.'

  def __init__(self, *ops:Operator) -> None:
    self.ops = ops
    self.level = -1


  @property
  def sub_refs(self) -> Iterator[RuleRef]:
    for op in self.ops:
      yield from op.sub_refs



class Left(Group):
  'Left-associative operator precedence group.'
  level_bump = 1



class Right(Group):
  'Right-associative operator precedence group.'



class Precedence(Rule):
  'An operator precedence rule, consisting of groups of operators.'
  type_desc = 'precedence rule'

  def __init__(self, leaves:Union[RuleRef,Iterable[RuleRef]], *groups:Group, drop:Iterable[str]=(),
   transform:UnaryTransform=unary_identity) -> None:
    # Keep track of the distinction between subs that came from leaves vs groups.
    # This allows us to catenate them all together to sub_refs, so they all get correctly linked,
    # and then get the linked leaves back via the leaves property implemented below.
    if isinstance(leaves, (str,Rule)): leaves = (leaves,)
    self.leaf_refs = tuple(leaves)
    self.group_refs = tuple(ref for g in groups for ref in g.sub_refs)
    self.name = ''
    self.sub_refs = self.leaf_refs + self.group_refs
    self.heads = ()
    self.drop = frozenset(iter_str(drop))
    self.transform = transform
    self.groups = groups
    self.head_table:Dict[TokenKind,Rule] = {}
    self.tail_table:Dict[TokenKind,Tuple[Group,Operator]] = {}


  def token_kinds(self) -> Iterable[str]:
    for group in self.groups:
      for op in group.ops:
        try: yield from op.kinds
        except _AllLeafKinds: pass # Adjacency ops do not directly reference any token.


  def head_subs(self) -> Iterable[Rule]:
    return iter(self.subs[:len(self.leaf_refs)]) # Only the leaves can be heads.


  def compile(self, parser:'Parser') -> None:
    for head in self.heads:
      matching_subs = [s for s in self.subs if head in s.heads]
      assert matching_subs
      if len(matching_subs) > 1:
        raise Parser.DefinitionError(f'{self} contains ambiguous primaries for head token {head!r}:\n',
          *indent_lines(str(s) for s in matching_subs))
      self.head_table[head] = matching_subs[0]

    for i, group in enumerate(self.groups):
      group.level = i*10 # Multiplying by ten lets us fudge the level by one to achieve right-associativity.
      for op in group.ops:
        try: kinds = op.kinds
        except _AllLeafKinds: kinds = tuple(self.head_table)
        for kind in kinds:
          try: existing = self.tail_table[kind]
          except KeyError: pass
          else: raise Parser.DefinitionError(f'{self} contains ambiguous operators for token {kind!r}:\n', existing, op)
          self.tail_table[kind] = (group, op)


  def parse(self, parent:Rule, source:Source, token:Token, buffer:Buffer[Token]) -> Any:
    syn = self.parse_level(parent, source, token, buffer, 0)
    return self.transform(source, token, syn)


  def parse_level(self, parent:Rule, source:Source, token:Token, buffer:Buffer[Token], level:int) -> Any:
    left = self.parse_leaf(parent, source, token, buffer)
    while True:
      op_token = next(buffer)
      while op_token.kind in self.drop:
        op_token = next(buffer)
      try:
        group, op = self.tail_table[op_token.kind]
      except KeyError:
        break # op_token is not an operator.
      if group.level < level: break # This operator is at a lower precedence.
      left = op.parse_right(parent, source, left, op_token, buffer, self.parse_level, level=group.level+group.level_bump)
    # op_token is either not an operator, or of a lower precedence level.
    buffer.push(op_token) # Put it back.
    return left


  def parse_leaf(self, parent:Rule, source:Source, token:Token, buffer:Buffer[Token]) -> Any:
    start = token
    while token.kind in self.drop:
      token = next(buffer)
    try: sub = self.head_table[token.kind]
    except KeyError: pass
    else: return self.parse_sub(sub, source, start, token, buffer)
    exp = self.name or f'any of {self.subs_desc}'
    raise ParseError(source, token, f'{parent} expects {exp}; received {token.kind}.')


class SubParser(Rule):

  def __init__(self, parser:'Parser', rule_name:str, transform:UnaryTransform=unary_identity) -> None:
    self.name = ''
    self.sub_refs = ()
    self.heads = parser.rules[rule_name].heads
    self.parser = parser
    self.rule = parser.rules[rule_name]
    self.transform = transform

  def parse(self, parent:Rule, source:Source, token:Token, buffer:Buffer[Token]) -> Any:
    sub_res = self.rule.parse(self, source, token, buffer)
    return self.transform(source, token, sub_res)


class Parser:
  '''
  '''

  class DefinitionError(Exception):
    def __init__(self, *msgs:Any) -> None:
      super().__init__(''.join(str(msg) for msg in msgs))


  def __init__(self, lexer:Lexer, rules:Dict[RuleName,Rule], literals:Iterable[TokenKind]=(), drop:Iterable[TokenKind]=()) -> None:
    self.lexer = lexer
    self.rules = rules
    self.literals = frozenset(iter_str(literals))
    self.drop = frozenset(iter_str(drop))
    self.module_name = get_caller_module_name()
    self._struct_types:Dict[str,Type] = {}

    for name, rule in rules.items():
      validate_name(name)
      rule.name = name

    # Link rule graph. Note: this creates reference cycles which are dissolved by __del__.

    def link_sub_ref(rule:RuleRef) -> Rule:
      if isinstance(rule, Rule): return rule
      if not isinstance(rule, str): raise Parser.DefinitionError(f'subrule must be a Rule or string reference: {rule!r}')
      try: return rules[rule]
      except KeyError: pass
      if rule in self.lexer.kinds: # Add the implied Atom rule.
        atom = Atom(rule)
        atom.name = rule
        self.rules[rule] = atom
        return atom
      raise Parser.DefinitionError(f'nonexistent rule: {rule!r}')

    def link(rule:Rule) -> Iterable[Rule]:
      assert not rule.subs
      if rule.sub_refs:
        rule.subs = tuple(link_sub_ref(s) for s in rule.sub_refs)
      assert len(rule.subs) == len(rule.sub_refs)
      return rule.subs

    self.nodes = visit_nodes(self.rules.values(), link) # All of the rule nodes.

    for rule in self.nodes:
      # Validate token references.
      for token_kind in rule.token_kinds():
        if token_kind not in lexer.kinds:
          raise Parser.DefinitionError(f'{rule} refers to nonexistent token kind: {token_kind}')
      # Fill out rule heads.
      if not rule.heads:
        heads = set(rule.compile_heads())
        try: heads.remove(_sentinel_kind)
        except KeyError: pass
        rule.heads = tuple(sorted(heads))
        assert rule.heads

    # Compile.
    for rule in self.nodes:
      rule.compile(parser=self)

    self.types:Immutable[Type] = Immutable(**self._struct_types)


  def __del__(self) -> None:
    for rule in self.rules.values():
      rule.__dict__.clear() # Break all references between rules.


  def _mk_struct_transform(self, name:str, subs:Tuple[Rule,...]) -> StructTransform:

    includes = [sub.name not in self.literals for sub in subs]
    field_names = [sub.name for sub in subs] # Allow empty names as they are; namedtuple will rename them.
    included_field_names = tuple(n for n, include in zip(field_names, includes) if include)

    if not any(includes):
      raise Parser.DefinitionError(f'struct rule {name} contains all literal fields: {field_names}; default transformer is degenerate.')

    if includes.count(True) == 1: # No need for a struct; just extract the interesting child element.
      i = includes.index(True)
      def single_transform(source:Source, token:Token, fields:List[Any]) -> Any: return fields[i]
      return single_transform

    struct_type = self._mk_struct_type(name, fields=included_field_names)

    def transform(source:Source, token:Token, fields:List[Any]) -> Any:
      return struct_type(*(f for f, include in zip(fields, includes) if include))

    return transform


  def _mk_struct_type(self, name:str, fields:Tuple[str,...]) -> Type:
    type_name = name.capitalize() or '_'.join(n.capitalize() for n in fields)
    try: existing = self._struct_types[type_name]
    except KeyError: pass
    else:
      if existing._fields != fields:
        raise Parser.DefinitionError(f'conflicting fields for synthesized struct type {name}:\n  {existing._fields}\n  {fields}')
      return existing
    struct_type = namedtuple(type_name, fields, rename=True, module=self.module_name or '?') # type: ignore
    self._struct_types[type_name] = struct_type
    return struct_type


  def make_buffer(self, source:Source, dbg_tokens:bool) -> Buffer[Token]:
    stream = self.lexer.lex(source, drop=self.drop, eot=True)
    if dbg_tokens:
      stream = tee_to_err(stream, label='Parser dbg_tokens')
    return Buffer(stream)


  def parse(self, rule_name:RuleName, source:Source, ignore_excess=False, dbg_tokens=False) -> Any:
    rule = self.rules[rule_name]
    buffer = self.make_buffer(source, dbg_tokens)
    token = next(buffer)
    result = rule.parse(parent=rule, source=source, token=token, buffer=buffer) # Top rule is passed as its own parent.
    excess_token = next(buffer) # Must exist because end_of_text cannot be consumed by a legal parser.
    if not ignore_excess and excess_token.kind != 'end_of_text':
      raise ExcessToken(source, excess_token, f'excess token: {excess_token.mode_kind}.')
    return result


  def parse_or_fail(self, rule_name:RuleName, source:Source, ignore_excess=False, dbg_tokens=False) -> Any:
    try: return self.parse(rule_name=rule_name, source=source, ignore_excess=ignore_excess, dbg_tokens=dbg_tokens)
    except ParseError as e: e.fail()


  def parse_all(self, rule_name:RuleName, source:Source, dbg_tokens=False) -> Iterator[Any]:
    rule = self.rules[rule_name]
    buffer = self.make_buffer(source, dbg_tokens=dbg_tokens)
    while True:
      token = next(buffer)
      if token.kind == 'end_of_text': return
      yield rule.parse(parent=rule, source=source, token=token, buffer=buffer) # Top rule is passed as its own parent.



def validate_name(name:Any) -> str:
  if not isinstance(name, str):
    raise Parser.DefinitionError(f'name is not a string: {name!r}')
  if not valid_name_re.fullmatch(name):
    raise Parser.DefinitionError(f'invalid name: {name!r}')
  if name in reserved_names:
    raise Parser.DefinitionError(f'name is reserved: {name!r}')
  return name



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from argparse import ArgumentParser
from html.parser import HTMLParser
from sys import stdin
from typing import List, Optional, Tuple


def main() -> None:
  arg_parser = ArgumentParser('Extract a portion of an HTML document.')
  arg_parser.add_argument('-id',help='The `id` of the DOM element to extract.')
  arg_parser.add_argument('path', nargs='?', help='path to the HTML document (defaults to stdin).')
  args = arg_parser.parse_args()
  path = args.path

  try: file = open(path) if path is not None else stdin
  except FileNotFoundError as e: exit(f'file not found: {e.filename}')
  parser = Parser(path=file.name, id=args.id, lines=list(file))
  parser.extract()


Pos = Tuple[int, int]


class Parser(HTMLParser):

  def __init__(self, path: str, id=str, lines=List[str]) -> None:
    super().__init__(convert_charrefs=True)
    self.path = path
    self.id = id
    self.lines = lines
    self.stack: List[Tuple[Pos, str]] = []
    self.extract_start_pos: Optional[Pos] = None

  def extract(self) -> None:
    for line in self.lines:
      self.feed(line)
    self.close()
    if self.extract_start_pos: exit('specified element was found but unterminated.')
    else: exit('specified element was not found.')

  def handle_starttag(self, tag: str, attrs: List[Tuple[str,Optional[str]]]):
    self.stack.append((self.pos, tag))
    d = dict(attrs)
    if d.get('id') == self.id:
      self.extract_start_pos = self.pos

  def handle_endtag(self, tag: str) -> None:
    if self.stack and self.stack[-1][1] == tag:
      p, t = self.stack.pop()
      if p == self.extract_start_pos:
        self.print_range(p, self.pos)
        exit(0)
      return
    self.msg(f'unmatched closing tag: {tag}')
    self.msg(f'in: {" ".join(t for _, t in self.stack)}')
    for i in reversed(range(len(self.stack))):
      pi, ti = self.stack[i]
      if ti == tag: # found match.
        for p, t in self.stack[i+1:]:
          self.msg(f'note: ignoring open `{t}` here', pos=p)
        self.msg(f'note: could match here', pos=pi)
        return

  def print_range(self, start_pos: Pos, end_pos: Pos) -> None:
    sl, sc = start_pos
    el, ec = end_pos
    if sl == el:
      l = self.lines[sl]
      print(l[sc:ec], end='')
    else:
      print(self.lines[sl][sc:], end='')
      for line in self.lines[sl+1:el]: print(line, end='')
      print(self.lines[el][:ec], end='')
    # Need to print end tag, which comes after `ec`.
    last = self.lines[el]
    close_pos = last.index('>', ec)
    print(last[ec:close_pos+1])

  @property
  def pos(self) -> Pos:
    line1, col0 = self.getpos()
    return (line1-1, col0)

  def msg(self, msg: str, pos:Pos=None) -> None:
    if pos is None: pos = self.pos
    print(f'{self.path}:{pos[0]+1}:{pos[1]+1}: {msg}')


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import re
from typing import Any, Generic, Iterator, TypeVar


_T = TypeVar('_T')

_setattr = object.__setattr__


class Immutable(Generic[_T]):
  'Untyped immutable object.'

  def __init__(self, obj:Any=None, **kw:Any) -> None:
    if obj is not None:
      vars(self).update(obj)
    for k, v in kw.items():
      _setattr(self, k, v)

  def __repr__(self) -> str:
    args = ', '.join(f'{k if _is_identifier(k) else repr(k)}={v!r}' for k, v in vars(self).items())
    return f'{type(self).__name__}({args})'

  def __getattr__(self, key:str) -> _T:
    return vars(self)[key] # type: ignore

  def __setattr__(self, name:str, val:_T) -> None:
    raise AttributeError('Immutable instance attributes are readonly')

  def __delattr__(self, name:str) -> None:
    raise AttributeError('Immutable instance attributes are readonly')

  def __getitem__(self, key:str) -> _T:
    return vars(self)[key] # type: ignore

  def __hash__(self) -> int:
    h = 0
    for p in vars(self).items():
      h ^= hash(p)
    return h

  def __eq__(self, other:Any) -> bool:
    return type(self) == type(other) and vars(self) == vars(other)

  def __iter__(self) -> Iterator[_T]:
    return iter(vars(self).items()) # type: ignore


_identifier_re = re.compile(r'(?!\d)\w+')
_is_identifier = _identifier_re.fullmatch



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.


'''
The tag_parse module provides a means for quickly constructing parsers for simple languages,
where each node of the syntax tree is either a lexical token
(as recognized by a regular expression)
or else a branch node delimited by distinct open and close tokens surrounding any child nodes.
Examples of such languages include HTML
(by treating the attributes in an opening tag as an opaque part of the token)
and Lisp-style S-Expressions.
The leaves of the syntax tree are simply strings;
the branches are TagTree instances, which are subclasses of tuple.
'''

import re

from itertools import chain
from typing import Callable, Dict, Match, Iterable, Iterator, List, Sequence, Tuple, Union
from .ansi import TXT_R, TXT_Y, RST_TXT, sgr
from .buffer import Buffer


class TagRule():
  def __init__(self, open_pattern: str, close_pattern: str, open_close_tokens_match_fn: Callable=None) -> None:
    self.open_pattern = open_pattern
    self.close_pattern = close_pattern
    self.open_close_tokens_match_fn = open_close_tokens_match_fn


class TagParser():

  def __init__(self, leaf_patterns: Sequence[str], tag_rules: Sequence[TagRule]) -> None:
    # parser works by wrapping each start and end pattern in a capture group.
    # the group itself is not used,
    # but the match group index indicates which open or close pattern matched.
    open_choices = ['({})'.format(r.open_pattern) for r in tag_rules]
    close_choices = ['({})'.format(r.close_pattern) for r in tag_rules]
    lexer_pattern = '|'.join(chain(open_choices, close_choices, leaf_patterns))
    self.lexer = re.compile(lexer_pattern)
    self.open_close_tokens_match_fns = [r.open_close_tokens_match_fn for r in tag_rules]
    self.last_open_index = len(tag_rules)


  def _parse(self, leaf_replacements: Dict[str, str], text: str, match_stream: Buffer[Match],
    pos: int, depth: int, subs: List, close_pred: Callable, parent_close_pred: Callable) -> Tuple['TagTree', int]:

    def append_leaf(leaf: str) -> None:
      subs.append(leaf_replacements.get(leaf, leaf))

    def flush_leaf(pos: int, end_index: int) -> None:
      leaf_text = text[pos:end_index]
      if leaf_text:
        append_leaf(leaf_text)

    for match in match_stream:
      flush_leaf(pos, match.start())
      pos = match.end()
      token = match.group()
      match_index = match.lastindex # only start and end tokens have a group.
      if match_index is None: # leaf token.
        append_leaf(token)
      elif match_index <= self.last_open_index: # found a start token (groups are 1-indexed).
        open_close_tokens_match_fn = self.open_close_tokens_match_fns[match_index - 1]
        exp_index = match_index + self.last_open_index
        def child_close_pred(close_match_index, close_token):
          return (close_match_index == exp_index
           and (open_close_tokens_match_fn is None
             or open_close_tokens_match_fn(token, close_token)))
        sub, pos = self._parse(leaf_replacements, text, match_stream, pos, depth + 1,
          subs=[token], close_pred=child_close_pred, parent_close_pred=close_pred)
        subs.append(sub)
      elif close_pred(match_index, token):
        subs.append(token)
        return TagTree(*subs), pos
      elif parent_close_pred(match_index, token): # parent end; missing end token.
        match_stream.push(match) # put the parent end token back into the stream.
        return TagTreeUnterminated(*subs), pos
      else: # unexpected end token.
        subs.append(TagTreeUnexpected(token))
    # end.
    flush_leaf(pos, len(text))
    return TagTreeRoot(*subs) if depth == 0 else TagTreeUnterminated(*subs), len(text)


  def parse(self, text: str, leaf_replacements: Dict[str, str]=None) -> 'Node':
    match_stream = Buffer(self.lexer.finditer(text))
    res, pos = self._parse(leaf_replacements or {}, text, match_stream, pos=0, depth=0,
      subs=[], close_pred=lambda i, t: False, parent_close_pred=lambda i, t: False)
    return res


Node = Union[str,'TagTree']


class TagTree(tuple):
  '''
  TagParser AST node.
  The str() value of the node is equal to the original text.
  '''

  def __new__(cls, *args: Node) -> 'TagTree':
    for el in args:
      if not isinstance(el, (str, TagTree)): raise ValueError(el)
    return super().__new__(cls, args) # type: ignore

  def __repr__(self) -> str:
    return '{}({})'.format(type(self).__name__, ', '.join(repr(el) for el in self))

  def __str__(self) -> str:
    return ''.join(self.walk_all())

  def __getitem__(self, key) -> Node: # type: ignore
    if isinstance(key, slice):
      return type(self)(*super().__getitem__(key)) # create a TagTree as the slice.
    return super().__getitem__(key) # type: ignore

  class_label = 'Tag'
  ansi_color = ''
  ansi_reset = ''

  @property
  def is_flawed(self) -> bool:
    return isinstance(self, TagTreeFlawed) or self.has_flawed_els

  @property
  def has_flawed_els(self) -> bool:
    return any(isinstance(el, TagTree) and el.is_flawed for el in self)

  @property
  def contents(self) -> Iterable[Node]:
    if len(self) < 2: raise ValueError('bad TagTree: {!r}'.format(self))
    return self[1:len(self)-1] # omit start and end tag.


  def walk_all(self) -> Iterator[str]:
    for el in self:
      if isinstance(el, str):
        yield el
      else:
        yield from el.walk_all()


  def walk_contents(self) -> Iterator[str]:
    for el in self.contents:
      if isinstance(el, str):
        yield el
      else:
        yield from el.walk_contents()


  def walk_branches(self, should_enter_tag_fn=lambda tag: True) -> Iterator['TagTree']:
    for el in self:
      if isinstance(el, TagTree):
        yield el
        if should_enter_tag_fn(el[0]):
          yield from el.walk_branches(should_enter_tag_fn=should_enter_tag_fn)


  def _structured_desc(self, res: List[str], depth: int) -> None:
    'multiline indented description helper.'
    if self.ansi_color: res.append(self.ansi_color)
    res.append(self.class_label)
    res.append(':')
    if self.ansi_reset: res.append(self.ansi_reset)
    d = depth + 1
    nest_spacer = '\n' + ('  ' * d)
    spacer = ' ' # the spacer to use for string tokens.
    for el in self:
      if isinstance(el, str):
        if spacer: res.append(spacer)
        res.append(repr(el))
        spacer = ''
      else:
        res.append(nest_spacer)
        el._structured_desc(res, d)
        spacer = nest_spacer

  def structured_desc(self, depth=0) -> str:
    res: List[str] = []
    self._structured_desc(res, depth)
    return ''.join(res)


class TagTreeRoot(TagTree):
  class_label = 'Root'

  @property
  def contents(self) -> Sequence[Node]:
    return self[:] # no tags at root level.


class TagTreeFlawed(TagTree):
  'abstract parent of the various flaw types.'
  pass


class TagTreeUnexpected(TagTreeFlawed):
  'unexpected trees consist solely of an unpaired closing tag.'

  class_label = 'Unexpected'
  ansi_color = TXT_R
  ansi_reset = sgr(RST_TXT)

  def __new__(cls, *args) -> 'TagTreeUnexpected':
    assert len(args) == 1
    return super().__new__(cls, *args) # type: ignore

  @property
  def contents(self) -> Sequence[Node]:
    assert len(self) == 1
    return ()


class TagTreeUnterminated(TagTreeFlawed):
  'unterminated trees are missing a closing tag.'

  class_label = 'Unterminated'
  ansi_color = TXT_Y
  ansi_reset = sgr(RST_TXT)

  @property
  def contents(self) -> Iterable[Node]:
    return self[1:len(self)] # omit the start tag only.






from difflib import SequenceMatcher
from typing import Any, Callable, Dict, List, Sequence, Tuple, TypeVar

from .patience import Alignment, Diff, align_patience


_T = TypeVar('_T')


def calc_diff(seq_a:Sequence[_T], seq_b:Sequence[_T], algorithm='patience', **kwargs:Any):
  alignment = alignment_fns[algorithm](seq_a, seq_b, **kwargs)
  improve_alignment(seq_a, seq_b, alignment)
  return diff_for_alignment(alignment, len(seq_a), len(seq_b))


def diff_for_alignment(alignment:Alignment, len_a:int, len_b:int) -> Diff:
  'Create a diff (list of range pairs) from an alignment (list of index pairs).'
  diff:Diff = []
  start_a = 0
  start_b = 0
  match_len = 0
  def flush(a:int, b:int) -> None:
    nonlocal start_a, start_b, match_len
    stop_a = start_a + match_len
    stop_b = start_b + match_len
    if match_len:
      diff.append((range(start_a, stop_a), range(start_b, stop_b)))
    if stop_a < a: diff.append((range(stop_a, a), range(stop_b, stop_b)))
    if stop_b < b: diff.append((range(a, a), range(stop_b, b)))
    start_a = a
    start_b = b

  for a, b in alignment:
    if start_a + match_len < a or start_b + match_len < b:
      flush(a, b)
      match_len = 1
    else:
      match_len += 1

  flush(len_a, len_b)
  return diff


def improve_alignment(seq_a:Sequence[_T], seq_b:Sequence[_T], alignment:Alignment) -> None:
  if not alignment: return
  prev_a, prev_b = alignment[0]
  for i, (a, b) in enumerate(alignment):
    last_a = a - 1
    last_b = b - 1
    if prev_a < last_a and prev_b  == last_b: # rem range.
      first_a = prev_a + 1 # index of first unaligned element.
      if seq_a[first_a] == seq_a[a]: # Change from bottom-aligned to top-aligned.
        a = first_a
        alignment[i] = (a, b)
    elif prev_a == last_a and prev_b < last_b: # add range.
      first_b = prev_b + 1
      if seq_b[first_b] == seq_b[b]:
        b = first_b
        alignment[i] = (a, b)
    prev_a = a
    prev_b = b


def validate_diff(seq_a:Sequence[_T], seq_b:Sequence[_T], diff:Diff, allow_empty=False) -> None:
  i_a = 0
  i_b = 0
  for r_a, r_b in diff:
    assert allow_empty or r_a or r_b
    if r_a:
      assert r_a.step == 1
      assert i_a == r_a.start
      i_a = r_a.stop
    if r_b:
      assert r_b.step == 1
      assert i_b == r_b.start
      i_b = r_b.stop
    if r_a and r_b:
      assert len(r_a) == len(r_b)
      for a, b in zip(r_a, r_b):
        assert seq_a[a] == seq_b[b]
  assert i_a == len(seq_a)
  assert i_b == len(seq_b)


def align_difflib(seq_a:Sequence[_T], seq_b:Sequence[_T], isjunk:Callable[[str], bool]=str.isspace, **kwargs) -> Alignment:
  'Diff using Python difflib.SequenceMatcher.'
  blocks:List[Tuple[int, int, int]] = SequenceMatcher(isjunk=isjunk, a=seq_a, b=seq_b, **kwargs).get_matching_blocks() # type: ignore
  assert blocks[-1] == (len(seq_a), len(seq_b), 0)
  return [(i_a+j, i_b+j) for (i_a, i_b, l) in blocks for j in range(l)]


alignment_fns: Dict[str, Callable[..., Alignment]] = {
  'difflib': align_difflib,
  'patience': align_patience,
}



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Any

from html5_parser import parse

from ..loader import FileOrPath, binary_file_for
from ..html import HtmlNode


def load_html(file_or_path:FileOrPath, encoding:str=None, **kwargs:Any) -> Any:
  with binary_file_for(file_or_path) as file:
    data = file.read()
  html = parse(data, transport_encoding=encoding, return_root=True, **kwargs)
  if 'treebuilder' in kwargs: return html

  # If none of the html5_parser `treebuilder` options was supplied,
  # then it will use the fast `lxml` option by default.
  # Transform the resulting etree into a generic dictionary tree.
  return HtmlNode.from_etree(html)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from http import HTTPStatus
from shlex import quote as sh_quote
from typing import Any, Dict
from urllib.parse import urlencode, urlparse
from random import uniform as rand_uniform
from time import sleep
from .fs import make_dirs, move_file, path_dir, path_exists, path_ext, path_join
from .io import errSL
from .path_encode import path_for_url
from .task import runCO


class HTTPError(Exception):
  def __init__(self, msg:str, curl_code:int=0, status_code:int=-1) -> None:
    super().__init__(msg)
    self.curl_code = curl_code
    self.status_code = status_code


def fetch(url:str, cache_path:str=None, params:Dict[str, str]={}, headers:Dict[str, str]={}, expected_status_code=200, timeout=30, delay=0, delay_range=0, spoof_ua=False) -> str:
  "Fetch the data at `url` and save it to a path in the '_fetch' directory derived from the URL."
  if params:
    if '?' in url: raise ValueError("params specified but url already contains '?'")
    url += '?' + urlencode(params)
  if not cache_path:
    cache_path = path_for_url(url)
  path = path_join('_fetch', cache_path)
  path_tmp = path_join('_fetch/tmp', cache_path)
  if not path_exists(path, follow=False):
    cmd = ['curl', url, '--write-out', '%{http_code}', '--output', path_tmp]
    if spoof_ua:
      h = spoofing_headers()
      h.update(headers) # any explicit headers override the spoofing values.
      headers = h
    for k, v in headers.items():
      cmd.extend(('--header', f'{k}: {v}'))
    make_dirs(path_dir(path_tmp))
    errSL('fetch:', *[sh_quote(word) for word in cmd])
    curl_code, output = runCO(cmd)
    if curl_code != 0:
      raise HTTPError(f'curl failed with code: {curl_code}', curl_code=curl_code)
      # TODO: copy the error code explanations from `man curl`? Or parse them on the fly?
    try:
      status_code = int(output)
      status = HTTPStatus(status_code)
    except ValueError as e:
      raise HTTPError(f'curl returned strange HTTP code: {repr(output)}') from e
    if status_code != expected_status_code:
      raise HTTPError(msg=f'fetch failed with HTTP code: {status.value}: {status.phrase}; {status.description}.',
        status_code=status_code)
    make_dirs(path_dir(path))
    move_file(path_tmp, path)
    sleep_min = delay - delay_range * 0.5
    sleep_max = delay + delay_range * 0.5
    sleep_time = rand_uniform(sleep_min, sleep_max)
    if sleep_time > 0:
      sleep(sleep_time)
  return path


def load_url(url:str, ext:str=None, cache_path:str=None, params:Dict[str, str]={}, headers:Dict[str, str]={}, expected_status_code=200, timeout=30, delay=0, delay_range=0, spoof_ua=False, **kwargs:Any) -> Any:
  'Fetch the data at `url` and then load using `muck.load`.'
  from .loader import load
  if ext is None:
    if cache_path:
      ext = path_ext(cache_path)
    else:
      # extract the extension from the url path;
      # we cannot leave it to load because it sees the encoded path,
      # which may have url parameters/query/fragment.
      parts = urlparse(url)
      ext = path_ext(parts.path) # TODO: support compound extensions, e.g. 'tar.gz'.
  path = fetch(url, cache_path=cache_path, params=params, headers=headers, expected_status_code=expected_status_code,
    timeout=timeout, delay=delay, delay_range=delay_range, spoof_ua=spoof_ua)
  return load(path, ext=ext, **kwargs)


def spoofing_headers() -> Dict[str, str]:
  # Headers that Safari sent at one point. Not sure how up-to-date these ought to be.
  # TODO: allow imitating other browsers?
  return {
    'DNT': '1',
    'Upgrade-Insecure-Requests': '1',
    'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
    'Cache-Control': 'max-age=0',
    'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_2) AppleWebKit/602.3.12 (KHTML, like Gecko) Version/10.0.2 Safari/602.3.12',
  }



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
File open/load dispatch by file extension.
'''

from io import BufferedReader, BytesIO, TextIOWrapper
from typing import IO, Any, BinaryIO, Callable, Dict, Iterable, List, Set, TextIO, Union


__all__ = [
  'add_loader',
  'FileOrPath',
  'LoaderException',
  'load',
  'LoadFn',
]

FileOrPath = Union[IO, str]
LoadFn = Callable[..., Any]


def load(file_or_path:FileOrPath, ext:str=None, **kwargs:Any) -> Any:
  '''
  Select an appropriate loader based on the file extension, or `ext` if specified.

  If a loader is found, then it is called, passing `kwargs`.

  If no loader is found, raises KeyError.
  '''
  if ext is None:
    if isinstance(file_or_path, str):
      ext = _path_cmpd_ext(file_or_path)
    else:
      try: name = file_or_path.name
      except AttributeError as e:
        raise ValueError(f'load: no `ext` specified and file does not have `name` attribute: {file_or_path}') from e
      if not isinstance(name, str):
        raise ValueError(f'load: no `ext` specified and file has non-string `name` attribute: {name!r}')
      ext = _path_cmpd_ext(name)
  dispatch_ext = _last_ext(ext)
  try: load_fn = _loaders[dispatch_ext]
  except KeyError as e:
    raise ValueError(f'load: extension {dispatch_ext!r} does not match any available loader: {file_or_path!r}') from e
  try: return load_fn(file_or_path, ext=ext, **kwargs)
  except Exception as e:
    raise LoaderException(f'load failed: {file_or_path!r}') from e


class LoaderException(Exception): pass



def text_file_for(f:FileOrPath, **kwargs:Any) -> TextIO:
  if isinstance(f, TextIOWrapper): return f
  if isinstance(f, str): return open(f, 'r', **kwargs)
  try: return TextIOWrapper(f, **kwargs)
  except Exception as e:
    raise ValueError(f'load: required text file or path; received {f!r}') from e


def binary_file_for(f:FileOrPath) -> BinaryIO:
  if isinstance(f, BufferedReader): return f # type: ignore # 'Statement is unreachable'.
  #^ Not sure about the reachability error above. Has to do with the relationship between abstract IO and concrete BufferedReader.
  if isinstance(f, str): return open(f, 'rb')
  try: return BufferedReader(f) # type: ignore
  except Exception as e:
    raise ValueError(f'load: required binary file or path; received {f!r}') from e


def add_loader(ext:str, fn:LoadFn, _dflt=False) -> None:
  '''
  Register a loader function, which will be called by `load` for matching `ext`.
  `_dflt` is used to mark the default loaders as such so that they can be overridden without triggering an error.
  '''
  if ext and not ext.startswith('.'):
    raise ValueError(f"file extension does not start with '.': {ext!r}")
  try: prev_loader = _loaders[ext]
  except KeyError: pass
  else:
    if prev_loader not in _dflt_loaders:
      raise Exception(f'add_loader: extension previously registered: {ext!r}; loader: {prev_loader!r}')
  _loaders[ext] = fn
  if _dflt:
    _dflt_loaders.add(fn)


_loaders:Dict[str,LoadFn] = {}
_dflt_loaders:Set[LoadFn] = set()


def load_archive(f:FileOrPath, single_name=None, single_ext=None, **kwargs:Any) -> Any:
  from .archive import Archive
  archive = Archive(binary_file_for(f))
  if single_name is None and single_ext is None:
    if kwargs:
      raise ValueError('load_archive: `single_name` or `single_ext` not specified; no other options should be set')
    return archive
  # load single file.
  match_exact = (single_name is not None)
  for file in archive: # type: ignore
    if match_exact:
      if file.name != single_name: continue
    else:
      if not file.name.endswith(single_ext): continue
    return load(file, ext=single_ext, **kwargs)
  raise LookupError(f'load_archive: could not find specified {"single_name" if match_exact else "single_ext"} in archive: {single_name!r}; archive.file_names: {archive.file_names}')


def load_binary(f:FileOrPath, ext:str, **kwargs:Any) -> BinaryIO:
  assert not kwargs, kwargs # To allow for opening as bytes or text with any encoding.
  return binary_file_for(f)


#def load_brotli(f:FileOrPath, ext:str, **kwargs:Any) -> Any:
#  from brotli import decompress
#  with binary_file_for(f) as f:
#    data = decompress(f.read())
#  sub_ext = _sub_ext(ext)
#  from gzip import GzipFile
#  df = GzipFile(mode='rb', fileobj=binary_file_for(f))
#  return load(df, ext=sub_ext, **kwargs) # type: ignore


def load_csv(f:FileOrPath, ext:str, encoding:str=None, **kwargs:Any) -> Iterable[List[str]]:
  from .csv import load_csv as _load_csv
  return _load_csv(text_file_for(f, newline='', encoding=encoding), **kwargs)


def load_html(f:FileOrPath, ext:str, encoding:str=None, **kwargs:Any) -> Any:
  from .html.loader import load_html as _load_html
  bf = binary_file_for(f)
  return _load_html(bf, encoding=encoding, **kwargs)


def load_gz(f:FileOrPath, ext:str, **kwargs:Any) -> Any:
  sub_ext = _sub_ext(ext)
  if sub_ext.endswith('.tar'): # load_archive handles compressed tar stream faster.
    return load_archive(f, **kwargs)
  from gzip import GzipFile
  df = GzipFile(mode='rb', fileobj=binary_file_for(f))
  return load(df, ext=sub_ext, **kwargs) # type: ignore


def load_json(file_or_path:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from .json import load_json as _load_json
  return _load_json(text_file_for(file_or_path), **kwargs)

def load_jsonl(file_or_path:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from .json import load_jsonl  as _load_jsonl
  return _load_jsonl(text_file_for(file_or_path), **kwargs)

def load_jsons(file_or_path:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from .json import load_jsons as _load_jsons
  return _load_jsons(text_file_for(file_or_path), **kwargs)


def load_lzfse(file_or_path:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from lzfse import decompress # type: ignore
  sub_ext = _sub_ext(ext)
  with binary_file_for(file_or_path) as f:
    df = BytesIO(f.read())
  return load(df, ext=sub_ext, **kwargs)


def load_msgpack(file_or_path:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from .msgpack import load_msgpack as _load_msgpack
  return _load_msgpack(binary_file_for(file_or_path), **kwargs)


def load_msgpacks(file_or_path:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from .msgpack import load_msgpacks as _load_msgpacks
  return _load_msgpacks(binary_file_for(file_or_path), **kwargs)


def load_pickle(file_or_path:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from .pickle import load_pickle as _load_pickle
  return _load_pickle(binary_file_for(file_or_path), **kwargs)


def load_plist(file_or_path:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from plistlib import load as _load_plist
  return _load_plist(binary_file_for(file_or_path), **kwargs)


def load_eon(file_or_path:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from .eon import parse_eon
  with text_file_for(file_or_path) as f:
    text = f.read()
    return parse_eon(path=f.name, text=text, **kwargs)


def load_pyl(f:FileOrPath, ext:str, **kwargs:Any) -> Any:
  'Load a python literal AST file (Python equivalent of JSON).'
  from ast import literal_eval
  return literal_eval(text_file_for(f, **kwargs).read())


def load_sqlite(f:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from .sqlite import Connection
  if not isinstance(f, str):
    raise ValueError(f'load: sqlite files can only be opened with a string path; received {f!r}')
  return Connection(f'file:{f}?mode=ro', uri=True, **kwargs) # type: ignore # Open in read-only mode.


def load_txt(f:FileOrPath, ext:str, clip_ends=False, **kwargs:Any) -> Iterable[str]:
  f = text_file_for(f, **kwargs)
  if clip_ends: return (line.rstrip('\n\r') for line in f)
  return f


def load_xls(f:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from openpyxl import load_workbook # type: ignore
  if isinstance(f, str):
    return load_workbook(filename=f, **kwargs)
  else:
    raise Exception(f'load_xls cannot load from an already open file handle: {f}')


def load_xz(f:FileOrPath, ext:str, **kwargs:Any) -> Any:
  sub_ext = _sub_ext(ext)
  if sub_ext.endswith('.tar'): # load_archive handles compressed stream faster.
    # TODO: investigate whether a multithreaded subprocess and/or pipeline between two processes could be faster.
    return load_archive(f, **kwargs)
  from lzma import LZMAFile
  d = LZMAFile(binary_file_for(f))
  #b = BufferedReader(d) # type: ignore
  return load(d, ext=sub_ext, **kwargs)


def load_yaml(f:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from yaml import load as _load_yaml, SafeLoader
  loader = kwargs.pop('Loader', SafeLoader)
  return _load_yaml(binary_file_for(f), Loader=loader, **kwargs) # type: ignore


def load_zst(f:FileOrPath, ext:str, **kwargs:Any) -> Any:
  from zstandard import ZstdDecompressor # type: ignore
  decompressor = ZstdDecompressor()
  d = decompressor.stream_reader(binary_file_for(f))
  return load(d, ext=_sub_ext(ext), **kwargs)


add_loader('',          load_binary,    _dflt=True)
add_loader('.css',      load_txt,       _dflt=True)
add_loader('.csv',      load_csv,       _dflt=True)
add_loader('.gz',       load_gz,        _dflt=True)
add_loader('.eon',      load_eon,       _dflt=True)
add_loader('.html',     load_html,      _dflt=True)
add_loader('.json',     load_json,      _dflt=True)
add_loader('.jsonl',    load_jsonl,     _dflt=True)
add_loader('.jsons',    load_jsons,     _dflt=True)
add_loader('.lzfse',    load_lzfse,     _dflt=True)
add_loader('.msgpack',  load_msgpack,   _dflt=True)
add_loader('.msgpacks', load_msgpacks,  _dflt=True)
add_loader('.pickle',   load_pickle,    _dflt=True)
add_loader('.plist',    load_plist,     _dflt=True)
add_loader('.pyl',      load_pyl,       _dflt=True)
add_loader('.sqlite',   load_sqlite,    _dflt=True)
add_loader('.sqlite3',  load_sqlite,    _dflt=True)
add_loader('.tar',      load_archive,   _dflt=True)
add_loader('.txt',      load_txt,       _dflt=True)
add_loader('.xls',      load_xls,       _dflt=True)
add_loader('.xlsx',     load_xls,       _dflt=True)
add_loader('.xz',       load_xz,        _dflt=True)
add_loader('.yaml',     load_yaml,      _dflt=True)
add_loader('.zip',      load_archive,   _dflt=True)
add_loader('.zst',      load_zst,       _dflt=True)


# Compound path utilities.

def _path_cmpd_ext(path:str) -> str:
  start = 1 if path.startswith('.') else 0 # Account for hidden files.
  try: idx = path.index('.', start)
  except ValueError: return ''
  else: return path[idx:]


def _last_ext(cmpd_ext:str) -> str:
  try: idx = cmpd_ext.rindex('.')
  except ValueError: return cmpd_ext
  else: return cmpd_ext[idx:]


def _sub_ext(cmpd_ext:str) -> str:
  try: idx = cmpd_ext.rindex('.')
  except ValueError: return cmpd_ext
  else: return cmpd_ext[:idx]


def main() -> None:
  from sys import argv
  from .io import outD
  from .parse import ParseError

  args = argv[1:] or ['/dev/stdin']
  for path in args:
    try: obj = load(path)
    except ParseError as e: e.fail()
    outD(path, obj)


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import os as _os
import re as _re
import shutil as _shutil
import stat as _stat
import time as _time
from os import DirEntry, get_exec_path as _get_exec_path, mkdir as _mkdir, scandir as _scandir
from os.path import expanduser as _expanduser, realpath as _realpath
from typing import (IO, Any, Callable, Dict, FrozenSet, Iterable, Iterator, List, Optional, Tuple)

from .clonefile import clone
from .filestatus import (file_ctime, file_inode, file_mtime, file_mtime_or_zero, file_permissions, file_size, file_stat,
  file_status, is_dir, is_file, is_file_executable_by_owner, is_link, is_link_to_dir, is_link_to_file, is_mount, path_exists)

from .path import (MixedAbsoluteAndRelativePathsError, Path, PathOrFd, abs_or_norm_path, abs_path, is_path_abs, norm_path,
  path_descendants, path_dir, path_ext, path_join, path_name, rel_path, split_dir_name, str_path)

from .util import memoize


class PathAlreadyExists(Exception): pass
class PathHasNoDirError(Exception): pass


def add_file_execute_permissions(path:PathOrFd, *, follow:bool) -> None:
  old_perms = file_permissions(path, follow=follow)
  new_perms = old_perms | _stat.S_IXUSR | _stat.S_IXGRP | _stat.S_IXOTH
  _os.chmod(path, new_perms, follow_symlinks=follow)


def change_dir(path:PathOrFd) -> None: _os.chdir(path)


def clone_or_hardlink(src:str, dst:str, follow_symlinks=True, preserve_owner=True) -> None:
  clone(src=src, dst=dst, follow_symlinks=follow_symlinks, preserve_owner=preserve_owner, fallback=None) # TODO


def clone_or_symlink(src:str, dst:str, follow_symlinks=True, preserve_owner=True) -> None:
  clone(src=src, dst=dst, follow_symlinks=follow_symlinks, preserve_owner=preserve_owner, fallback=None) # TODO


def copy_eagerly(src:str, dst:str, follow_symlinks:bool=True, preserve_owner:bool=True) -> None:
  # TODO: implementation that more precisely matches semantics of APFS clonefile.
  # TODO: test on other platforms, figure out appropriate cross platform semantics. In particular, look at BTRFS and ZFS.
  _shutil.copytree(src=src, dst=dst)


def copy_path(src:str, dst:str, overwrite:bool=True, create_dirs:bool=False, follow_symlinks:bool=True,
 preserve_owner:bool=True) -> None:
  if overwrite and path_exists(dst, follow=False):
    remove_path(dst)
  if create_dirs:
    make_parent_dirs(dst)
  clone(src=src, dst=dst, follow_symlinks=follow_symlinks, preserve_owner=preserve_owner, fallback=copy_eagerly)


def copy_to_dir(src:str, dst:str,  overwrite:bool=True, create_dirs:bool=False, follow_symlinks:bool=True,
 preserve_owner:bool=True) -> None:
  if create_dirs:
    make_dirs(dst)
  return copy_path(src=src, dst=path_join(dst, path_name(src)),
    overwrite=overwrite, create_dirs=False, follow_symlinks=follow_symlinks, preserve_owner=preserve_owner)


def expand_user(path:Path) -> str: return _expanduser(str_path(path))


default_project_signifiers: Tuple[str, ...] = (
  '.git',
  '.project-root',
  'Package.swift',
  'craft.eon',
  'setup.py',
)

def find_project_dir(start_dir:Path='.', top:Optional[Path]=None, include_top=False,
 project_signifiers:Iterable[str]=default_project_signifiers) -> Optional[str]:
  '''
  find a project root directory, as denoted by the presence of a file/directory in `project_signifiers`.
  By default, stops before reaching the user's home directory.
  If a signifier string contains any of '?*+^$[]', then it is treated as a regular expression.
  See default_project_signifiers.
  '''
  signifier_re = _re.compile('|'.join(f'({p if any(q in p for q in "?*+^$[]") else _re.escape(p)})' for p in project_signifiers))
  start_dir = abs_path(start_dir)
  if top is None:
    top = home_dir()
    if not start_dir.startswith(top):
      top = '/'
  else:
    top = abs_path(top)
  for path in walk_dirs_up(start_dir, top=top, include_top=include_top):
    for name in list_dir(path, hidden=True):
      if signifier_re.fullmatch(name):
        return path
  return None


def home_dir() -> str: return _expanduser('~')


def is_python_file(path:Path, always_read=False) -> bool:
  '''
  Guess if a file is a python file, based first on path extension, or if that is not present on shebang line.
  TODO: support more than just '#!/usr/bin/env python'
  TODO: support zip archives?
  '''
  if not always_read:
    ext = path_ext(path)
    if ext: return ext == '.py'
  try:
    with open(path, 'rb') as f:
      expected = b'#!/usr/bin/env python'
      head = f.read(len(expected))
      return bool(head == expected)
  except (FileNotFoundError, IsADirectoryError): return False


def list_dir(path:PathOrFd, exts:Iterable[str]=(), hidden=False) -> List[str]:
  '''
  Return a list of the names in the directory at `path`,
  optionally filtering by extensions in `exts`, and the `hidden` flag (defaults to False, excluding names beginning with '.').
  '''
  exts = normalize_exts(exts)
  names = sorted(_os.listdir(path))
  #^ Sort is necessary for some file systems, e.g. APFS.
  if not exts and hidden: return names # no filtering necessary.
  return [n for n in names if name_has_any_ext(n, exts) and (hidden or not n.startswith('.'))]


def list_dir_paths(path:Path, exts:Iterable[str]=(), hidden=False) -> List[str]:
  return [path_join(path, name) for name in list_dir(path, exts=exts, hidden=hidden)]


def make_dir(path:Path) -> None:
  return _mkdir(path)


def make_dirs(path:Path, mode=0o777, exist_ok=True) -> None:
  '''
  Like os.makedirs, except:
  * uses `mode` to make all intermediate directories.
  * checks more carefully that existing paths are directories, raising NotADirectoryError.
  '''
  dir, name = split_dir_name(norm_path(path))
  # Recursively make the parent dirs.
  if not name:
    assert dir == '/'
  elif dir:
    ds = file_status(dir, follow=True)
    if ds is None: # Parent dir does not exist; try to create it.
      try: make_dirs(dir, mode=mode, exist_ok=exist_ok)
      except FileExistsError as e: # Another thread or process just created it.
        if not is_dir(dir, follow=True):
          raise NotADirectoryError(dir) from e
    elif not ds.is_dir:
      raise NotADirectoryError(dir)
  # Make the specified dir.
  try: _mkdir(path, mode)
  except OSError as e:
    # Cannot rely on checking for EEXIST, since the operating system
    # could give priority to other errors like EACCES or EROFS
    if not path_exists(path, follow=False): raise # Perhaps some other OSError.
    if not exist_ok: raise PathAlreadyExists(path) from e
    if not is_dir(path, follow=True): raise NotADirectoryError(path) from e
    return # The directory already exists.


def make_parent_dirs(path:Path, mode=0o777, exist_ok=True) -> None:
  dir = path_dir(path)
  if dir: make_dirs(dir, exist_ok=exist_ok)


def make_link(orig:Path, *, link:Path, absolute=False, allow_nonexistent=False, overwrite=False, create_dirs=False,
 perms:Optional[int]=None) -> None:
  if perms is not None: raise NotImplementedError # TODO
  abs_orig = abs_path(orig)
  if abs_orig == abs_path(link): raise ValueError(f'cannot create link to self: orig: {orig!r}; link: {link!r}')
  if not allow_nonexistent and not path_exists(orig, follow=True):
    raise FileNotFoundError(orig)
  if absolute:
    _orig = abs_orig
  else:
    _orig = rel_path(orig, start=path_dir(link))
  if create_dirs: make_parent_dirs(link)
  if path_exists(link, follow=False):
    if overwrite:
      remove_file(link)
    else:
      raise FileExistsError(link)
  return _os.symlink(_orig, link)


def move_file(path:Path, to:str, overwrite=False, create_dirs=False) -> None:
  if not overwrite and path_exists(to, follow=False):
    raise Exception('destination path already exists: {}'.format(to))
  if create_dirs: make_parent_dirs(path)
  _os.replace(path, to)


def name_has_any_ext(name:str, exts:FrozenSet[str]) -> bool:
  '''
  Returns True if `exts` is empty or any element of `exts` is a strict suffix of `path`.
  '''
  if name == '': raise ValueError(name) # Poorly defined for the empty string.
  if not exts: return True
  for ext in exts:
    if name != ext and name.endswith(ext): return True
  return False


def normalize_exts(exts:Iterable[str]) -> FrozenSet[str]:
  if isinstance(exts, str):
    return frozenset((exts,))
  for ext in exts:
    if not isinstance(ext, str): raise TypeError(ext)
    if ext and not ext.startswith('.'): raise ValueError(ext)
  return frozenset(exts)


def open_new(path:Path, create_dirs:bool=True, **open_args) -> IO[Any]:
  if path_exists(path, follow=False):
    raise PathAlreadyExists(path)
  if create_dirs: make_parent_dirs(path)
  return open(path, 'w', **open_args)


@memoize
def path_for_cmd(cmd: str) -> Optional[str]:
  for dir in _get_exec_path():
    try: entries = _scandir(dir)
    except FileNotFoundError: continue # directory in PATH might not exist.
    for entry in entries:
      if entry.name == cmd and entry.is_file: return path_join(dir, cmd)
  return None


def product_needs_update(product=PathOrFd, source=PathOrFd) -> bool: # type: ignore
  return file_mtime_or_zero(product, follow=True) <= file_mtime(source, follow=True)


read_link = _os.readlink


def real_path(path:Path) -> str: return _realpath(str_path(path))


def remove_dir(path:Path) -> None:
  remove_dir_contents(path, hidden=True)
  _os.rmdir(path)


def remove_dir_contents(path:Path, hidden=False) -> None:
  for n in list_dir_paths(path, hidden=hidden):
    remove_path(n)


def remove_dir_contents_if_exists(path:Path, hidden=False) -> None:
  if path_exists(path, follow=True): remove_dir_contents(path, hidden=hidden)


def remove_file(path:Path) -> None: _os.remove(path)


def remove_file_if_exists(path:Path) -> None:
  if path_exists(path, follow=False):
    remove_file(path)


def remove_empty_dir(path:Path) -> None: _os.rmdir(path)

def remove_empty_dirs(path:Path) -> None: _os.removedirs(path)


def remove_path(path:Path) -> None:
  if is_dir(path, follow=False): remove_dir(path)
  else: remove_file(path)


def remove_path_if_exists(path:Path) -> None:
  if path_exists(path, follow=False):
    remove_path(path)


def scan_dir(path:Path, exts:Iterable[str]=(), hidden=False) -> List[DirEntry]:
  exts = normalize_exts(exts)
  entries = sorted(_os.scandir(path), key=lambda e: e.name)
  if not exts and hidden: return entries
  return [e for e in entries if name_has_any_ext(e.name, exts) and (hidden or not e.name.startswith('.'))]


def set_mtime(path:PathOrFd, mtime:Optional[float]) -> None:
  '''
  Update the access and modification times of the file at `path`.
  The access time is always updated to the current time;
  `mtime` defaults to the current time.
  '''
  _os.utime(path, None if mtime is None else (_time.time(), mtime))


def touch_path(path:Path, mode=0o666) -> None:
  fd = _os.open(path, flags=_os.O_CREAT|_os.O_APPEND, mode=mode)
  try: _os.utime(fd)
  finally: _os.close(fd)


def walk_dirs(*paths:Path, make_abs=False, include_hidden=False, file_exts:Iterable[str]=()) -> Iterator[str]:
  return walk_paths(*paths, make_abs=make_abs, yield_files=False, yield_dirs=True,
    include_hidden=include_hidden, file_exts=file_exts)


def walk_dirs_and_files(*dir_paths:Path, make_abs=False, include_hidden=False, file_exts:Iterable[str]=(),
 files_as_paths=False) -> Iterator[Tuple[str, List[str]]]:
  '''
  yield (dir_path, files) pairs.
  files is an array of either names (default) or paths, depending on the files_as_paths option.
  '''
  file_exts = normalize_exts(file_exts)
  for raw_path in dir_paths:
    dir_path = abs_or_norm_path(raw_path, make_abs)
    if not dir_path.endswith('/'): dir_path += '/'
    yield from _walk_dirs_and_files(dir_path, include_hidden, file_exts, files_as_paths)


def _walk_dirs_and_files(dir_path:str, include_hidden:bool, file_exts:FrozenSet[str], files_as_paths:bool) -> Iterator[Tuple[str, List[str]]]:
  sub_dirs = []
  files = []
  assert dir_path.endswith('/')
  for name in list_dir(dir_path, hidden=include_hidden):
    path = dir_path + name
    if is_dir(path, follow=True):
      sub_dirs.append(path + '/')
    elif name_has_any_ext(name, file_exts):
      files.append(path if files_as_paths else name)
  yield (dir_path, files)
  for sub_dir in sub_dirs:
    yield from _walk_dirs_and_files(sub_dir, include_hidden, file_exts, files_as_paths)


def walk_dirs_up(path:Path, top:Path, include_top=True) -> Iterable[str]:
  if is_path_abs(path) ^ is_path_abs(top):
    raise MixedAbsoluteAndRelativePathsError((path, top))
  if is_dir(path, follow=True):
    dir_path = path
  else:
    dir_path = path_dir(path)
    if not dir_path:
      raise PathHasNoDirError(path)
  return reversed(path_descendants(top, dir_path))


def walk_files(*paths:Path, make_abs=False, include_hidden=False, file_exts:Iterable[str]=()) -> Iterator[str]:
  return walk_paths(*paths, make_abs=make_abs, yield_files=True, yield_dirs=False,
    include_hidden=include_hidden, file_exts=file_exts)


def walk_paths(*paths:Path, make_abs=False, yield_files=True, yield_dirs=True, include_hidden=False,
  file_exts:Iterable[str]=(), pass_dash=True) -> Iterator[str]:
  '''
  Generate file and/or dir paths, optionally filtering hidden names and/or by file extension.
  Treats `-` as a special symbol for stdin, and returns it unaltered and unfiltered as a special case.
  This special case can be turned off with `pass_dash=False`.
  '''
  file_exts = normalize_exts(file_exts)
  for raw_path in paths: # TODO: sort?
    if pass_dash and yield_files and raw_path == '-': # Special case to indicate stdin.
      yield '-'
      continue
    path = abs_or_norm_path(raw_path, make_abs)
    if is_dir(path, follow=True):
      yield from _walk_paths_rec(path + '/', yield_files, yield_dirs, include_hidden, file_exts)
    elif not path_exists(path, follow=True):
      raise FileNotFoundError(path)
    elif yield_files and name_has_any_ext(path_name(path), file_exts):
      yield path


def _walk_paths_rec(dir_path:str, yield_files:bool, yield_dirs:bool, include_hidden:bool, file_exts:FrozenSet[str]) -> Iterator[str]:
  'yield paths; directory paths are distinguished by trailing slash.'
  assert dir_path.endswith('/')
  if yield_dirs:
    yield dir_path
  try: names = list_dir(dir_path, hidden=include_hidden)
  except FileNotFoundError: return # It existed a moment ago; caller probably deleted the directory.
  for name in names:
    path = path_join(dir_path, name)
    if is_dir(path, follow=True):
      yield from _walk_paths_rec(path + '/', yield_files, yield_dirs, include_hidden, file_exts)
    elif yield_files and name_has_any_ext(name, file_exts):
      yield path


class DirEntries:

  def __init__(self, exts:Iterable[str]=(), hidden=False, pred:Callable[[DirEntry],bool]=None) -> None:
    self.exts  = normalize_exts(exts)
    self.hidden = hidden
    self.pred = (lambda entry:True) if pred is None else pred
    self._entries:Dict[str,Tuple[DirEntry,...]] = {}

  def __getitem__(self, dir_path:str) -> Tuple[DirEntry,...]:
    try: return self._entries[dir_path]
    except KeyError: pass

    try: v = tuple(filter(self.pred, scan_dir(dir_path, exts=self.exts, hidden=self.hidden)))
    except OSError as e: raise KeyError(dir_path) from e
    self._entries[dir_path] = v
    return v

  def clear(self) -> None:
    self._entries.clear()

  def clear_dir(self, dir:str) -> None:
    try: del self._entries[dir]
    except KeyError: pass



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from abc import abstractmethod
from typing import Any, Callable, Counter, Dict, Optional, Protocol, Tuple, TypeVar, Union


_T = TypeVar('_T')

NoneType = type(None)
Opt = Optional


class Comparable(Protocol):
  # taken from https://www.python.org/dev/peps/pep-0484/.
  @abstractmethod
  def __lt__(self, other:Any) -> bool: ...

  @abstractmethod
  def __eq__(self, other:Any) -> bool: ...


def is_a(val:Any, T:Union[type, Tuple[type, ...]]) -> bool:
  '''
  Test if `val` is of `T`. Unlike `isinstance`,
  this function works with generic static types.
  '''
  try: return isinstance(val, T)
  except TypeError as e:
    if e.args[0] != 'Subscripted generics cannot be used with class and instance checks': raise
  RTT = T.__origin__ # type: ignore # The runtime type for the static type.
  args = T.__args__  # type: ignore
  try:
    predicate = _generic_type_predicates[RTT]
  except KeyError: # Not specialized.
    if issubclass(RTT, dict): # Two parameters.
      if issubclass(RTT, Counter): # Counter only has one type parameter.
        K = args[0]
        V = int # Note that Counters can have non-int values inserted.
      else:
        K, V = args
      return isinstance(val, RTT) and all(is_a(k, K) and is_a(v, V) for (k, v) in val.items())
    elif len(args) == 1: # Assume `T` is a single-parameter generic container.
      E = args[0]
      return isinstance(val, RTT) and all(is_a(el, E) for el in val)
    else:
      raise TypeError(f'{T} is not a single-parameter generic type; origin type: {RTT}')
  else: # Specialized.
    # Union is an extra strange case, because the origin type is not a runtime type either.
    return (RTT is Union or isinstance(val, RTT)) and predicate(val, args)


_Args = Tuple[Any, ...]

def _is_a_Tuple(v:Any, args:_Args) -> bool:
  if len(args) == 2 and args[1] is Ellipsis:
    E = args[0]
    return all(is_a(el, E) for el in v)
  else:
    return len(v) == len(args) and all(is_a(el, E) for (el, E) in zip(v, args))

_generic_type_predicates: Dict[Any, Callable[[Any, _Args], bool]] = {
  tuple: _is_a_Tuple,
  Union: lambda v, args: any(is_a(v, Member) for Member in args),
}


# Convenience type predicates.

def is_bool(val: Any) -> bool: return isinstance(val, bool)

def is_float(val: Any) -> bool: return isinstance(val, float)

def is_int(val: Any) -> bool: return isinstance(val, int)

def is_str(val: Any) -> bool: return isinstance(val, str)

def is_list(val: Any, of:Optional[type]=None) -> bool:
  return isinstance(val, list) and (of is None or all(isinstance(el, of) for el in val))

def is_set(val: Any, of:Optional[type]=None) -> bool:
  return isinstance(val, set) and (of is None or all(isinstance(el, of) for el in val))

def is_dict(val: Any, of:Optional[type]=None) -> bool:
  return isinstance(val, dict) and (of is None or all(isinstance(el, of) for el in val))

def is_tuple(val: Any, of:Optional[type]=None, length:Optional[int]=None) -> bool:
  return isinstance(val, tuple) and (length is None or length == len(val)) and (of is None or all(isinstance(el, of) for el in val))

def is_int_or_bool(val: Any) -> bool: return isinstance(val, (int, bool))

def is_list_of_str(val: Any) -> bool: return isinstance(val, list) and all(isinstance(el, str) for el in val)

def is_set_of_str(val: Any) -> bool: return isinstance(val, set) and all(isinstance(el, str) for el in val)

def is_tuple_of_str(val: Any, length:Optional[int]=None) -> bool:
  return is_tuple(val, of=str, length=length)

def is_dict_of_str(val: Any) -> bool:
  return isinstance(val, dict) and all(isinstance(k, str) and isinstance(v, str) for (k, v) in val.items())

def is_pair_of_str(val: Any) -> bool: return is_tuple(val, of=str, length=2)

def is_pair_of_int(val: Any) -> bool: return is_tuple(val, of=int, length=2)

def is_str_or_list(val: Any) -> bool: return is_str(val) or is_list_of_str(val)

def is_str_or_pair(val: Any) -> bool: return is_str(val) or is_pair_of_str(val)

def is_pos_int(val: Any) -> bool: return is_int(val) and bool(val > 0)


def req_type(obj: _T, expected: Union[type, Tuple[type, ...]]) -> _T:
  'Return `obj` if it is of `expected` type, or else raise a descriptive TypeError.'
  if not is_a(obj, expected):
    raise TypeError(f'expected type: {expected}; actual type: {type(obj)};\n  object: {obj!r}')
  return obj



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import ByteString


def _byte_index(alphabet:bytes, char:int) -> int:
  try: return alphabet.index(char)
  except ValueError: return 0xff

base62_alphabet = b'0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
base62_alphabet_inverse = bytes(_byte_index(base62_alphabet, c) for c in range(0x100))
assert len(base62_alphabet) == 62

base58_alphabet = b'123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
base58_alphabet_inverse = bytes(_byte_index(base58_alphabet, c) for c in range(0x100))
assert len(base58_alphabet) == 58


def _lep_int_from_bytes(val:ByteString) -> int:
  '''
  Create a (big) integer from the little-endian interpretation of the bytes,
  and then add a leading 1, which will act as a terminator when decoding.
  '''
  n = int.from_bytes(val, byteorder='little')
  return n + (1<<(len(val)*8))


def lep_encode(val:ByteString, alphabet:bytes) -> bytes:
  m = len(alphabet)
  res = bytearray()
  n = _lep_int_from_bytes(val)
  while n:
    n, r = divmod(n, m)
    res.append(alphabet[r])
  return bytes(res)

def lep_decode(val:ByteString, alphabet:bytes, alphabet_inverse:bytes) -> bytes:
  m = len(alphabet)
  n = 0
  for i, char in enumerate(val):
    a = alphabet_inverse[char]
    if a >= m: raise ValueError(val)
    n += (m**i) * a
  res = bytearray()
  while n > 1:
    n, r = divmod(n, 0x100)
    res.append(r)
  if n != 1: raise ValueError(val)
  return bytes(res)


def enc_lep62(val:ByteString) -> bytes:
  return lep_encode(val, alphabet=base62_alphabet)

def dec_lep62(val:ByteString) -> bytes:
  return lep_decode(val, alphabet=base62_alphabet, alphabet_inverse=base62_alphabet_inverse)


def enc_b32(val:int) -> str:
  if not isinstance(val, int): raise TypeError
  chars = []
  if val == 0: return '0'
  neg = (val < 0)
  if neg: val = -val
  while val > 0:
    digit = val & 0x1f # low five bits.
    val >>= 5
    chars.append(chr((0x30 if digit < 10 else 0x37) + digit))
  if neg: chars.append('-')
  return ''.join(reversed(chars))



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
HTML writer.
'''

from typing import Any
from ..xml.tree import EscapedStr, XmlAttrs, XmlWriter, add_opt_attrs
from ..xml.escape import esc_xml_attr, esc_xml_text
from ..svg import Svg, Style, Script
from .semantics import form_input_types


class HtmlWriter(XmlWriter):
  '''
  HtmlWriter is a ContextManager class for generating HTML.

  '''

  can_auto_close_tags = False # Unlike XML, HTML5 dictates that each tag type either be self-closing or not.

  def a(self, *children:Any, **attrs:Any) -> 'A': return self.child(A, *children, attrs=attrs)

  def br(self) -> None: self.leaf('br', attrs=None)

  def div(self, *children:Any, **attrs:Any) -> 'Div': return self.child(Div, *children, attrs=attrs)

  def img(self, *children:Any, **attrs:Any) -> 'Img': return self.child(Img, *children, attrs=attrs)

  def h1(self, *children:Any, **attrs:Any) -> 'H1': return self.child(H1, *children, attrs=attrs)
  def h2(self, *children:Any, **attrs:Any) -> 'H2': return self.child(H2, *children, attrs=attrs)
  def h3(self, *children:Any, **attrs:Any) -> 'H3': return self.child(H3, *children, attrs=attrs)
  def h4(self, *children:Any, **attrs:Any) -> 'H4': return self.child(H4, *children, attrs=attrs)
  def h5(self, *children:Any, **attrs:Any) -> 'H5': return self.child(H5, *children, attrs=attrs)
  def h6(self, *children:Any, **attrs:Any) -> 'H6': return self.child(H6, *children, attrs=attrs)

  def hr(self) -> None: self.leaf('hr', attrs=None)

  def meta(self, **attrs:Any) -> 'Meta': return self.child(Meta, attrs=attrs)

  def p(self, *children:Any, **attrs:Any) -> 'P': return self.child(P, *children, attrs=attrs)

  def pre(self, *children:Any, **attrs:Any) -> 'Pre': return self.child(Pre, *children, attrs=attrs)

  def span(self, *children:Any, **attrs:Any) -> 'Span': return self.child(Span, *children, attrs=attrs)

  def script(self, *children:Any, **attrs:Any) -> Script: return self.child(Script, *children, attrs=attrs)

  def style(self, *children:Any, **attrs:Any) -> Style: return self.child(Style, *children, attrs=attrs)

  def svg(self, *children:Any, **kwargs:Any) -> Svg:
    return self.child(Svg, *children, **kwargs)

  # Forms.

  def form(self, *children:Any, **attrs:Any) -> 'Form': return self.child(Form, *children, attrs=attrs)

  def input(self, **attrs:Any) -> None:
    # TODO: move to Input class.
    if attrs.get('type') not in form_input_types:
      raise Exception(f'bad HTML <input> type: {attrs.get("type")}')
    return self.leaf('input', attrs=attrs)

  def label(self, *children:Any, **attrs:Any) -> 'Label':
    return self.child(Label, *children, attrs=attrs)

  # Tables.

  # TODO: return XmlWriter subclasses that enforce correct permitted parent/child structures.

  def table(self, *children:Any, **attrs:Any) -> 'Table': return self.child(Table, *children, attrs=attrs)

  def caption(self, *children:Any, **attrs:Any) -> 'Caption': return self.child(Caption, *children, attrs=attrs)

  def thead(self, *children:Any, **attrs:Any) -> 'THead': return self.child(THead, *children, attrs=attrs)

  def tfoot(self, *children:Any, **attrs:Any) -> 'TFoot': return self.child(TFoot, *children, attrs=attrs)

  def td(self, *children:Any, **attrs:Any) -> 'TD': return self.child(TD, *children, attrs=attrs)

  def th(self, *children:Any, **attrs:Any) -> 'TH': return self.child(TH, *children, attrs=attrs)

  def tr(self, *children:Any, **attrs:Any) -> 'TR': return self.child(TR, *children, attrs=attrs)

  def title(self, *children:Any, **attrs:Any) -> 'Title': return self.child(Title, *children, attrs=attrs)


class Html(HtmlWriter):
  tag = 'html'

  def __init__(self, attrs:XmlAttrs=None) -> None:
    super().__init__(attrs=attrs)
    self.prefix = '<!DOCTYPE html>'
    self.head = self.child(Head, attrs=attrs)
    self.head.meta(charset='utf-8')
    self.body = self.child(Body, attrs=attrs)


class A(HtmlWriter):
  tag = 'a'

class Body(HtmlWriter):
  tag = 'body'

class Div(HtmlWriter):
  tag = 'div'

class Img(HtmlWriter):
  tag = 'img'
  is_self_closing = True

class Head(HtmlWriter):
  tag = 'head'

class H1(HtmlWriter):
  tag = 'h1'

class H2(HtmlWriter):
  tag='h2'

class H3(HtmlWriter):
  tag = 'h3'

class H4(HtmlWriter):
  tag = 'h4'

class H5(HtmlWriter):
  tag = 'h5'

class H6(HtmlWriter):
  tag = 'h6'

class P(HtmlWriter):
  tag = 'p'

class Pre(HtmlWriter):
  tag = 'pre'

class Form(HtmlWriter):
  tag = 'form'

class Input(HtmlWriter):
  tag = 'input'

class Label(HtmlWriter):
  tag = 'label'

class Meta(HtmlWriter):
  tag = 'meta'
  is_self_closing = True

class Span(HtmlWriter):
  tag = 'span'

class Table(HtmlWriter):
  tag = 'table'

class Caption(HtmlWriter):
  tag = 'caption'

class THead(HtmlWriter):
  tag = 'thead'

class TFoot(HtmlWriter):
  tag = 'tfoot'

class TD(HtmlWriter):
  tag = 'td'

class TH(HtmlWriter):
  tag = 'th'

class TR(HtmlWriter):
  tag = 'tr'

class Title(HtmlWriter):
  tag = 'title'




# Copyright 2011 George King. Permission to use this file is granted in license-gloss.txt.

r'''
TODO: register a signal handler for SIGWINCH to update sizes.
'''

import fcntl as _fcntl
import struct as _struct
from copy import deepcopy
from sys import stderr, stdout
from termios import (BRKINT, CS8, CSIZE, ECHO, ICANON, ICRNL, IEXTEN, INPCK, ISIG, ISTRIP, IXON, OPOST, PARENB, TCSADRAIN,
  TCSAFLUSH, TCSANOW, TIOCGWINSZ, VMIN, VTIME, tcgetattr, tcsetattr)

from .typing import OptBaseExc, OptTraceback, OptTypeBaseExc


def window_size(f=stdout):
  '''
  TODO: replace with shutil.get_terminal_size()?
  '''
  if not f.isatty():
    return (128, 0)
  try:
    cr = _struct.unpack('hh', _fcntl.ioctl(f, TIOCGWINSZ, b'xxxx')) # arg string length indicates length of return bytes
  except:
    print('pithy.term.window_size: ioctl failed', file=stderr)
    raise
  return int(cr[1]), int(cr[0])


# Indexes for termios list (see <termios.h>).
IFLAG = 0
OFLAG = 1
CFLAG = 2
LFLAG = 3
ISPEED = 4
OSPEED = 5
CC = 6


when_vals = (TCSANOW, TCSAFLUSH, TCSADRAIN)


class TermMode:
  '''
  A context manager for altering terminal modes.
  If no file descriptor is provided, it defaults to stdout.
  '''

  def __init__(self, fd=None, when=TCSAFLUSH, min_bytes=1, delay=0) -> None:
    assert when in when_vals, when
    if fd is None:
      fd = stdout.fileno()
    self.fd = fd
    self.when = when
    self.min_bytes = min_bytes
    self.original_attrs = tcgetattr(fd)
    self.attrs = deepcopy(self.original_attrs)
    self.vtime = 0
    if delay > 0:
      self.vtime = int(delay * 10)
      if self.vtime <= 0: raise ValueError(f'delay must be 0 or greater than 0.1s; received: {delay}')
    self.alter_attrs()

  def __enter__(self):
    tcsetattr(self.fd, self.when, self.attrs)

  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    tcsetattr(self.fd, self.when, self.original_attrs)

  def alter_attrs(self) -> None:
    raise NotImplementedError('TermMode must be subclassed.')


class CBreakMode(TermMode):

  def alter_attrs(self) -> None:
    attrs = self.attrs
    attrs[IFLAG] &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON) # type: ignore
    attrs[OFLAG] &= ~(OPOST) # type: ignore
    attrs[CFLAG] &= ~(CSIZE | PARENB) # type: ignore
    attrs[CFLAG] |= CS8 # type: ignore
    attrs[LFLAG] &= ~(ECHO | ICANON | IEXTEN | ISIG) # type: ignore
    attrs[CC][VMIN] = self.min_bytes # type: ignore
    attrs[CC][VTIME] = self.vtime # type: ignore


class RawMode(TermMode):

  def alter_attrs(self) -> None:
    attrs = self.attrs
    attrs[LFLAG] &= ~(ECHO | ICANON) # type: ignore
    attrs[CC][VMIN] = self.min_bytes # type: ignore
    attrs[CC][VTIME] = self.vtime # type: ignore


class SilentMode(TermMode):

  def alter_attrs(self) -> None:
    attrs = self.attrs
    attrs[LFLAG] &= ~(ECHO) # type: ignore



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
Xml type.
'''

from typing import Any, Iterable, List, Union
from xml.etree.ElementTree import Element
from xml.sax.saxutils import escape as _escape_text, quoteattr as _escape_attr

from ..exceptions import MultipleMatchesError, NoMatchError
from ..markup import Mu, MuAttrItem, MuAttrs, MuChild, _Mu


#from ..desc import repr_lim
#from ..exceptions import DeleteNode, FlattenNode, MultipleMatchesError, NoMatchError, ConflictingValues
#from .escape import fmt_attr_items

XmlAttrs = MuAttrs
XmlAttrItem = MuAttrItem

XmlChild = Union[str,'Xml']
XmlChildren = List[XmlChild]


class XmlNode(Mu):
  '''
  Xml subclass adds XML-specific details to Mu base type.
  '''

  def esc_attr_val(self, val:str) -> str: return _escape_attr(val)

  def esc_text(self, text:str) -> str: return _escape_text(text) # Ignoring custom entities for now.


class Xml(Mu):
  '''
  Generic Xml node that contains its own tag.
  '''
  __slots__ = ('tag')

  def __init__(self:_Mu, *, tag, attrs:MuAttrs=None, ch:Iterable[MuChild]=(), cl:Iterable[str]=None,
   _orig:_Mu=None, _parent:'Mu'=None, **kw_attrs:Any) -> None:
    self.tag = tag
    super().__init__(tag=tag, attrs=attrs, ch=ch, cl=cl, _orig=_orig, _parent=_parent, **kw_attrs) # type: ignore



#!/usr/bin/env python3
# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import platform
from os import execv
from os.path import abspath as abs_path, join as path_join
from shlex import quote as sh_quote
from sys import argv, executable, stderr


def main() -> None:
  if len(argv) < 2: exit('usage: py-lldb [script] ...')
  if not executable: exit('py-lldb error: python executable path is not available.')
  exe = executable
  if platform.system() == 'Darwin':
    # Darwin's python3 binary is a wrapper that execs the actual binary.
    # Attempt to use this directly so that lldb debugs the real process.
    exe = abs_path(path_join(exe, '../../Resources/Python.app/Contents/MacOS/Python'))
  lldb = '/usr/bin/lldb'
  lldb_args = ['--batch', '--one-line', 'run', '--', exe, *argv[1:]]
  quoted = ' '.join(sh_quote(a) for a in [lldb, *lldb_args])
  print(f'py-lldb exec: {quoted}', file=stderr)
  execv(lldb, lldb_args)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from types import MappingProxyType
from typing import Any


def freeze(value: Any, dicts=True, lists=True, sets=True, bytearrays=True) -> Any:
  if dicts and isinstance(value, dict):
    return MappingProxyType({k: freeze(v) for k, v in value.items()})
  if lists and isinstance(value, list):
    return tuple(freeze(el) for el in value)
  if sets and isinstance(value, set):
    return frozenset(freeze(el) for el in value)
  if bytearrays and isinstance(value, bytearray):
    return bytes(value)
  return value



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from enum import Enum
from itertools import tee
from operator import le
from typing import (Any, Callable, Dict, FrozenSet, Hashable, Iterable, Iterator, List, Mapping, Optional, Set, Tuple, TypeVar,
  Union)

from .types import Comparable


_T = TypeVar('_T')
_K = TypeVar('_K', bound=Hashable)
_V = TypeVar('_V')
_C = TypeVar('_C', bound=Comparable)
_CK = TypeVar('_CK', bound=Comparable)


class MultipleElements(ValueError):
  'Raised when an iterable unexpectedly has multiple elements.'


class NoElements(KeyError):
  'Raised when an iterable unexpectedly has no elements.'


known_leaf_types = (bool, bytearray, bytes, complex, float, int, str, range, type(None), type(Ellipsis))
#^ Note: the only type that is strictly necessary is `str`, because its element type is `str` which causes infinite recursion.
# It also makes sense not to iterate over byte values or range indices, as that is probably not the intent of a generic iteration.
# The remaining members are a speculative optimization.


def is_sorted(iterable: Iterable, cmp=le) -> bool:
  'Test whether `iterable` is sorted according to `cmp` (default = `operator.le`).'
  a, b = tee(iterable)
  next(b, None) # map works like zip: stops as soon as any input is exhausted. The default `None` is ignored here.
  return all(map(cmp, a, b))


def first_el(iterable: Iterable[_T]) -> _T:
  'Advance `iterable` and return the first element or raise `ValueError`.'
  for el in iterable: return el
  raise NoElements(iterable)


def single_el(iterable:Iterable[_T]) -> _T:
  first = init = object()
  for el in iterable:
    if first == init:
      first = el
    else:
      raise MultipleElements((first, el))
  if first == init: raise NoElements(iterable)
  return first # type: ignore


def iter_from(iterable: Iterable[_T], start: int) -> Iterator[_T]:
  'Return an iterator over `iterable` that skips elements up to `start` index.'
  it = iter(iterable)
  c = start
  while c > 0:
    try: next(it)
    except StopIteration: break
    c -= 1
  return it


def iter_unique(iterable: Iterable[_T]) -> Iterator[_T]:
  'Drop repeated elements, like unix `uniq`. TODO: rename to `iter_drop_repeated`.'
  prev:Any = object()
  for el in iterable:
    if el != prev:
      yield el
      prev = el


def iter_values(obj:Any) -> Iterator[Any]:
  if isinstance(obj, known_leaf_types): return
  if hasattr(obj, 'values'): # Treat as a mapping.
    yield from obj.values()
  else:
    try: it = iter(obj)
    except TypeError: return
    yield from it


def extent(iterable: Iterable[_C], key: Callable[[_C], _CK]=None, default: Optional[_C]=None) -> Tuple[_C, _C]:
  'Return the min and max.'
  it = iter(iterable)
  first = next(it) if default is None else next(it, default)
  l = first
  h = first
  if key is None:
    for el in it:
      if el < l:
        l = el
      if h < el:
        h = el
  else:
    kl = key(l)
    kh = key(h)
    for el in it:
      k = key(el)
      if k < kl:
        l = el
        kl = k
      if kh < k:
        h = el
        kh = k
  return (l, h)


def count_by_pred(iterable: Iterable[_T], pred: Callable[[_T], Any]) -> int:
  count = 0
  for el in iterable:
    if pred(el): count += 1
  return count


def closed_int_intervals(iterable: Iterable[int]) -> Iterable[Tuple[int, int]]:
  'Given `iterable` of integers, yield a sequence of closed intervals.'
  it = iter(iterable)
  try: first = next(it)
  except StopIteration: return
  if not isinstance(first, int):
    raise TypeError('closed_int_intervals requires a sequence of int elements; received first element: {!r}', first)
  interval = (first, first)
  for i in it:
    l, h = interval
    if i < h:
      raise ValueError('closed_int_intervals requires monotonically increasing elements')
    if i == h: continue
    if i == h + 1:
      interval = (l, i)
    else:
      yield interval
      interval = (i, i)
  yield interval


_RangeTypes = Union[int, range, Tuple[int, int]]
def int_tuple_ranges(iterable: Iterable[_RangeTypes]) -> Iterable[Tuple[int, int]]:
  'Given `iterable`, yield range pair tuples.'

  def pair_for_el(el: _RangeTypes) -> Tuple[int, int]:
    if isinstance(el, range): return (el.start, el.stop)
    if isinstance(el, int): return (el, el + 1)
    assert isinstance(el, tuple)
    if len(el) != 2: raise ValueError(el)
    return el

  it = iter(iterable)
  try: low, end = pair_for_el(next(it))
  except StopIteration: return
  for el in it:
    l, e = pair_for_el(el)
    if e < l: raise ValueError(el)
    if l < end: raise ValueError('seq_int_ranges requires monotonically increasing elements', end, l)
    if l == end:
      end = e
    else:
      yield (low, end)
      low = l
      end = e
  yield (low, end)


def filtermap_with_mapping(iterable: Iterable[_K], mapping: Mapping[_K, _V]) -> Iterable[_V]:
  'Map the values of `iterable` through the mapping, dropping any elements not in the mapping.'
  for el in iterable:
    try: yield mapping[el]
    except KeyError: pass


def fan_by_index_fn(iterable: Iterable[_T], index: Callable[[_T], int], min_len=0) -> List[List[_T]]:
  '''
  Fan out `iterable` into a list of lists, with a minimum length of `min_len`,
  according to the index returned by applying `index` to each element.
  '''
  l: List[List[_T]] = []
  while len(l) < min_len:
    l.append([])
  for el in iterable:
    i = int(index(el))
    if i < 0: raise IndexError(i)
    while len(l) <= i:
      l.append([])
    l[i].append(el)
  return l


def fan_by_pred(iterable: Iterable[_T], pred: Callable[[_T], bool]) -> Tuple[List[_T], List[_T]]:
  'Fan out `iterable` into a pair of lists by applying `pred` to each element.'
  fan: Tuple[List[_T], List[_T]] = ([], [])
  for el in iterable:
    if pred(el):
      fan[1].append(el)
    else:
      fan[0].append(el)
  return fan


def fan_by_key_fn(iterable: Iterable[_T], key: Callable[[_T], _K]) -> Dict[_K, List[_T]]:
  '''
  Fan out `iterable` into a dictionary by applying a function `key` that returns a group key for each element.
  returns a dictionary of lists.
  '''
  groups: Dict[_K, List[_T]] = {}
  for el in iterable:
    k = key(el)
    try:
      group = groups[k]
    except KeyError:
      group = []
      groups[k] = group
    group.append(el)
  return groups


def group_sorted_by_cmp(iterable: Iterable[_T], cmp: Callable[[_T, _T], bool]) -> List[List[_T]]:
  '''
  Group elements `iterable`, which must already be sorted,
  by applying the `comparison` predicate to each consecutive pair of elements.
  Consecutive elements for which the predicate returns truthy will be grouped together;
  a group is yielded whenever comparison fails.
  '''
  # TODO: convert to generator.
  it = iter(iterable)
  try:
    first = next(it)
  except StopIteration:
    return []
  # TODO: rename group.
  groups = []
  group = [first]
  prev = first
  for el in it:
    if cmp(prev, el):
      group.append(el)
    else:
      groups.append(group)
      group = [el]
    prev = el
  if group:
    groups.append(group)
  return groups


class OnHeadless(Enum):
  error, drop, keep = range(3)


def group_by_heads(iterable: Iterable[_T], is_head: Callable[[_T], bool], headless=OnHeadless.error, keep_heads=True) \
 -> Iterator[List[_T]]:
  '''
  Group elements of `iterable` by creating a new group every time the `is_head` predicate evaluates to true.
  If the first element of the stream is not a head, the behavior is specified by `headless`.
  '''
  it = iter(iterable)
  group: List[_T] = []
  while True: # consume all headless (leading tail) tokens.
    try: el = next(it)
    except StopIteration:
      if group: yield group # Headless.keep.
      return
    if is_head(el):
      if group: # Headless.keep.
        yield group
        group = []
      if keep_heads: group.append(el)
      break
    else: # leading tail element.
      if headless == OnHeadless.error: raise ValueError(el)
      if headless == OnHeadless.drop: continue
      if headless == OnHeadless.keep: group.append(el)
      else: raise TypeError(headless)
  for el in it:
    if is_head(el):
      yield group
      group = [el] if keep_heads else []
    else:
      group.append(el)
  if group: yield group


def set_from(iterables:Iterable[Iterable[_K]]) -> Set[_K]:
  s:Set[_K] = set()
  for el in iterables:
    s.update(el)
  return s


def frozenset_from(iterables:Iterable[Iterable[_K]]) -> FrozenSet[_K]:
  return frozenset(set_from(iterables))


def split_els(iterable:Iterable[_T], split=Callable[[_T], Optional[Tuple[_T, _T]]]) -> Iterator[_T]:
  '''
  Repeatedly split the current element using the `split` function until it returns None.
  '''
  for el in iterable:
    r = split(el)
    while r is not None:
      head, el = r
      yield head
      r = split(el)
    yield el


def split_by_preds(iterable: Iterable[_T], *preds: Callable[[_T], bool]) -> Iterable[Tuple[bool, List[_T]]]:
  '''
  Split the sequence whenever the sequence of predicates has consecutively matched.
  Each yielded chunk is a pair (is_split_seq, seq).
  For example:
  `split_by_preds('abcde', lambda el: el=='b', lambde el: el=='c')` yields:
  * (False, ['a'])
  * (True, ['b', 'c'])
  * (False, ['d'. 'e'])
  '''
  if not preds: raise ValueError('split_by_preds requires at least one predicate')
  l = len(preds)
  buffer: List[_T] = []
  for el in iterable:
    buffer.append(el)
    if len(buffer) >= l:
      tail = buffer[-l:]
      if all(p(e) for p, e in zip(preds, tail)):
        if len(buffer) > l:
          yield (False, buffer[:-l])
        yield (True, tail)
        buffer.clear()
  if buffer:
    yield (False, buffer)


def split_halves_by_pred(iterable:Iterable[_T], pred:Callable[[_T], bool]) -> Tuple[List[_T],List[_T]]:
  left:List[_T] = []
  right:List[_T] = []
  it = iter(iterable)
  for el in it:
    if pred(el):
      right.append(el)
      break
    else:
      left.append(el)
  right.extend(it)
  return left, right


def window_iter(iterable: Iterable[_T], width=2) -> Iterator[Tuple[_T, ...]]:
  'Yield tuples of the specified `width` (default 2), consisting of adjacent elements in `seq`.'
  # TODO: use tee? might be faster.
  assert width > 0
  buffer = []
  for el in iterable:
    buffer.append(el)
    if len(buffer) == width:
      yield tuple(buffer)
      del buffer[0]


def window_pairs(iterable: Iterable[_T], tail: Optional[_T]=None) -> Iterator[Tuple[_T, Optional[_T]]]:
  'Yield pairs of adjacent elements in `seq`, including a final pair consisting of the last element and `tail`.'
  it = iter(iterable)
  try: head = next(it)
  except StopIteration: return
  for el in it:
    yield (head, el)
    head = el
  yield (head, tail)


_PrefixTreeTerminator = TypeVar('_PrefixTreeTerminator', bound=Hashable)
PrefixTree = Dict[Union[_K, _PrefixTreeTerminator], Optional[Dict]] # mypy cannot handle recursive types.


def prefix_tree(iterables:Iterable[Iterable[_K]], terminator:_PrefixTreeTerminator=None, update:PrefixTree=None) -> PrefixTree:
  res:PrefixTree = {} if update is None else update
  for it in iterables:
    d = res
    for el in it:
      try:
        e = d[el]
        assert e is not None
        d = e
      except KeyError:
        e = {}
        d[el] = e
        d = e
    d[terminator] = None
  return res



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from bisect import bisect
from itertools import chain
from typing import Dict

from . import CodeRange, CodeRanges, abbreviated_planes, intersect_sorted_ranges, union_sorted_ranges
from .categories import unicode_categories, unicode_category_aliases
from .data_11_00 import blocks, category_ranges


def is_code_in_charset(code, charset:CodeRanges) -> bool:
  p = (code, code)
  i = bisect(charset, p)
  if i < len(charset):
    s, e = charset[i]
    if s <= code < e: return True
  return i > 0 and code < charset[i - 1][1]


unicode_charsets:Dict[str,CodeRanges] = {}
ascii_charsets:Dict[str,CodeRanges] = {}

def _gen_charsets() -> None:

  # categories.
  for cat in unicode_categories:
    if cat.subcategories:
      unicode_charsets[cat.key] = tuple(sorted(chain(*(category_ranges[k] for k in cat.subcategories))))
    else:
      unicode_charsets[cat.key] = category_ranges[cat.key]
  # add aliases in a second pass to reuse range values by reference.
  for name, cat in unicode_category_aliases.items():
    if name in unicode_charsets: continue
    unicode_charsets[name] = unicode_charsets[cat.key]

  # blocks.
  for k, r in blocks.items():
    unicode_charsets[k] = (r,)

  # planes.
  for k, plane in abbreviated_planes.items():
    unicode_charsets[k] = plane

  def add(name:str, abbr:str, *ranges:CodeRange) -> None:
    unicode_charsets[name] = tuple(ranges)
    if abbr: unicode_charsets[abbr] = unicode_charsets[name]

  # Ascii.
  add('Ascii', 'A', (0x00, 0x80))
  Ascii = unicode_charsets['Ascii']

  for cat in unicode_categories:
    ranges = tuple(intersect_sorted_ranges(Ascii, unicode_charsets[cat.name]))
    if not ranges: continue
    k = 'Ascii_' + cat.name
    add(k, 'A' + cat.key, *ranges)
    ascii_charsets[k] = unicode_charsets[k]

  # Control characters. Note that 'Cc' also includes 0x7F (DEL).
  add('Control_0', 'C0', (0x0000, 0x0020))
  add('Control_1', 'C1', (0x0080, 0x00A0))

  # Common numeric sets for programming languages.
  add('Binary',       'Bin',  (0x30, 0x32))
  add('Quaternary',   'Quat', (0x30, 0x34))
  add('Octal',        'Oct',  (0x30, 0x38))
  add('Decimal',      'Dec',  (0x30, 0x3A)) # remove as redundant with Ascii_Number?
  add('Hexadecimal',  'Hex',  (0x30, 0x3A), (0x41, 0x47), (0x61, 0x67))

  # Miscellaneous.

  add('Visible', 'V', *union_sorted_ranges(*[unicode_charsets[k] for k in ['L', 'M', 'N', 'P', 'S']]))
  add('Readable', 'R', *union_sorted_ranges(*[unicode_charsets[k] for k in ['L', 'M', 'N', 'P', 'S', 'Z']]))

  add('Ascii_Visible', 'AV', *intersect_sorted_ranges(Ascii, unicode_charsets['V']))
  add('Ascii_Readable', 'AR', *intersect_sorted_ranges(Ascii, unicode_charsets['R']))

_gen_charsets()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from bisect import bisect
from typing import Dict, List, Optional, Sequence, Tuple, TypeVar


_T = TypeVar('_T')

Alignment = List[Tuple[int, int]]
Diff = List[Tuple[range, range]]


def align_patience(seq_a:Sequence[_T], seq_b:Sequence[_T]) -> Alignment:
  '''
  Generate a sequence alignment using the Patience algorithm.
  '''
  alignment: Alignment = []
  _align(seq_a, seq_b, 0, 0, len(seq_a), len(seq_b), alignment)
  return alignment


def _align(seq_a:Sequence[_T], seq_b:Sequence[_T], pos_a:int, pos_b:int, end_a:int, end_b:int, alignment:List[Tuple[int, int]]) -> None:
  '''
  Find alignment between `seq_a` and `seq_b`.
  Results accumulate in `alignment` as (idx_a, idx_b) pairs.
  '''
  #print(f'\n{pos_a} {pos_b}; {end_a} {end_b}')
  assert pos_a <= end_a and pos_b <= end_b, f'{pos_a} <= {end_a}, {pos_b} <= {end_b}'

  while pos_a < end_a and pos_b < end_b and seq_a[pos_a] == seq_b[pos_b]:
    alignment.append((pos_a, pos_b))
    pos_a += 1
    pos_b += 1

  if pos_a == end_a or pos_b == end_b: return
  end_alignment = []
  while pos_a < end_a and pos_b < end_b and seq_a[end_a-1] == seq_b[end_b-1]:
    end_alignment.append((end_a-1, end_b-1))
    end_a -= 1
    end_b -= 1

  if pos_a == end_a or pos_b == end_b: # TODO: optimize for len < 3 case?
    alignment.extend(reversed(end_alignment))
    return

  assert pos_a < end_a and pos_b < end_b, f'{pos_a} < {end_a}, {pos_b} < {end_b}'
  unique_alignment = unique_lcs(seq_a[pos_a:end_a], seq_b[pos_b:end_b])
  if unique_alignment:
    prev_a = pos_a
    prev_b = pos_b
    for a, b in unique_alignment:
      a += pos_a
      b += pos_b
      assert seq_a[a] == seq_b[b]
      _align(seq_a, seq_b, prev_a, prev_b, a, b, alignment)
      alignment.append((a, b))
      prev_a = a + 1
      prev_b = b + 1
    _align(seq_a, seq_b, prev_a, prev_b, end_a, end_b, alignment)
  alignment.extend(reversed(end_alignment))


def unique_lcs(seq_a:Sequence[_T], seq_b:Sequence[_T]) -> List[Tuple[int, int]]:
  '''
  Find the longest common subset for unique elements.
  See http://en.wikipedia.org/wiki/Patience_sorting
  '''

  index_a = index_uniques(seq_a)
  index_b = index_uniques(seq_b)

  b_to_a:List[Optional[int]] = [None if index_b[el] is None else index_a.get(el) for el in seq_b]

  back_refs:List[Optional[int]] = [None] * len(seq_b)
  piles:List[int] = []
  lasts:List[int] = []
  pile_idx = 0
  for i_b, i_a in enumerate(b_to_a):
    if i_a is None: continue
    if piles and piles[-1] < i_a: # Optimization.
      pile_idx = len(piles)
    elif piles and piles[pile_idx] < i_a and (pile_idx == len(piles) - 1 or piles[pile_idx+1] > i_a): # Optimization.
      pile_idx += 1
    else:
      pile_idx = bisect(piles, i_a)
    if pile_idx: # Not the first pile; need a back_ref.
      back_refs[i_b] = lasts[pile_idx-1]
    if pile_idx < len(piles):
      piles[pile_idx] = i_a
      lasts[pile_idx] = i_b
    else:
      piles.append(i_a)
      lasts.append(i_b)

  if not piles: return []

  result:List[Tuple[int, int]] = []
  last:Optional[int] = lasts[-1]
  while last is not None:
    o_a = b_to_a[last]
    assert o_a is not None
    result.append((o_a, last))
    last = back_refs[last]
  result.reverse()
  return result


def index_uniques(seq:Sequence[_T]) -> Dict[_T, Optional[int]]:
  index:Dict[_T, Optional[int]] = {}
  for i, el in enumerate(seq):
    index[el] = None if el in index else i
  return index



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
XML escaping utilities.
'''

from functools import lru_cache
from html import escape as html_escape

from typing import Any, Iterable, Tuple, Optional, Dict, List


XmlAttrs = Optional[Dict[str,Any]]


class EscapedStr(str):
  'A `str` subclass that signifies to some receiver that it has already been properly escaped.'


def esc_xml_text(val:Any) -> str:
  'HTML-escape the string representation of `val`.'
  # TODO: add options to support whitespace escaping?
  return val if isinstance(val, EscapedStr) else html_escape(str(val), quote=False)


def esc_xml_attr(val:Any) -> str:
  'HTML-escape the string representation of `val`, including quote characters.'
  return val if isinstance(val, EscapedStr) else html_escape(str(val), quote=True)


@lru_cache(maxsize=1024, typed=True)
def esc_xml_attr_key(key:str) -> str:
  return html_escape(key.replace("_", "-"), quote=True)


def fmt_attrs(attrs:XmlAttrs, replaced_attrs:Dict[str,str]) -> str:
  'Format the `attrs` dict into XML key-value attributes.'
  if not attrs: return ''
  return fmt_attr_items(attrs.items(), replaced_attrs)


def fmt_attr_items(attr_items:Iterable[Tuple[str,Any]], replaced_attrs:Dict[str,str]) -> str:
  parts: List[str] = []
  for k, v in attr_items:
    k = replaced_attrs.get(k, k)
    if v is None: v = 'none'
    parts.append(f' {esc_xml_attr_key(k)}="{esc_xml_attr(v)}"')
  return ''.join(parts)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import re
from argparse import ArgumentParser
from itertools import groupby
from os import environ
from sys import stderr, stdout
from typing import Any, Dict, List, Match, Optional, Set, Tuple

from pithy.ansi import BG, BOLD, FILL, RST, RST_BOLD, RST_TXT, TXT, gray26, rgb6, sanitize_for_console, sgr

from ..diff import calc_diff


'''
same-same is a git diff highlighter.

To use it, add the following configuration to your .gitconfig:
[core]
  pager = same-same | LESSANSIENDCHARS=mK less --RAW-CONTROL-CHARS
[interactive]
  diffFilter = same-same -interactive | LESSANSIENDCHARS=mK less --RAW-CONTROL-CHARS
[diff]
  noprefix = true
  wsErrorHighlight = none # Intraline coloring creates trouble for same-same.

To disable, set 'SAME_SAME_OFF' in the shell environment.'
'''


class DiffLine:
  def __init__(self, kind:str, match:Match) -> None:
    self.kind = kind # The name from `diff_pat` named capture groups.
    self.match = match
    self.old_num = 0 # 1-indexed.
    self.new_num = 0 # ".
    self.chunk_idx = 0 # Positive for rem/add.
    self.is_src = False # True for ctx/rem/add.
    self.text = '' # Final text for ctx/rem/add.

  @property
  def raw_text(self) -> str:
    return self.match.string # type: ignore

  def set_text(self, key:str, clip:bool=False) -> None:
    text = self.match[key]
    if self.match['git_color'] or clip:
      text = clip_reset(text)
    self.text = text


def main() -> None:

  arg_parser = ArgumentParser(prog='same-same', description='Git diff filter.')
  arg_parser.add_argument('-interactive', action='store_true', help="Accommodate git's interactive mode.")
  args = arg_parser.parse_args()

  # Git can generate utf8-illegal sequences; ignore them.
  stdin = open(0, errors='replace')

  if 'SAME_SAME_OFF' in environ:
    for line in stdin:
      stdout.write(line)
    exit(0)

  dbg = ('SAME_SAME_DBG' in environ)

  # Break input into segments starting with 'diff' lines.
  # Note that the first segment might begin with any kind of line.
  buffer:List[DiffLine] = []

  def flush_buffer() -> None:
    nonlocal buffer
    if buffer:
      handle_file_lines(buffer, interactive=args.interactive)
      buffer = []

  try:
    for line in stdin:
      raw_text = line.rstrip('\n')
      match = diff_pat.match(raw_text)
      assert match is not None
      kind = match.lastgroup
      assert kind is not None, match
      if dbg:
        print(kind, ':', repr(raw_text))
        continue
      if kind in pass_kinds:
        flush_buffer()
        print(raw_text)
        continue
      if kind == 'diff':
        flush_buffer()
      buffer.append(DiffLine(kind, match))
    flush_buffer()
  except BrokenPipeError:
    stderr.close() # Prevents warning message.


def handle_file_lines(lines:List[DiffLine], interactive:bool) -> None:
  first = lines[0]
  kind = first.kind

  # If we are processing `git log --graph` then parsing will fail; detect and skip.
  if git_diff_graph_mode_pat.match(first.raw_text).end(): # type: ignore
    for line in lines: print(line.raw_text)
    return

  # Scan `lines` to build up diff structures.

  old_ctx_nums:Set[int] = set() # Line numbers of context lines.
  new_ctx_nums:Set[int] = set() # ".
  old_lines:Dict[int, DiffLine] = {} # Maps of line numbers to line structs.
  new_lines:Dict[int, DiffLine] = {} # ".
  old_uniques:Dict[str, Optional[int]] = {} # Maps unique line bodies to line numbers.
  new_uniques:Dict[str, Optional[int]] = {} # ".
  old_num = 0 # 1-indexed source line number.
  new_num = 0 # ".
  chunk_idx = 0 # Counter to differentiate chunks; becomes part of the groupby key.
  old_path = '<OLD_PATH>'
  new_path = '<NEW_PATH>'
  is_prev_add_rem = False
  is_loc_colored = False # Because git diff does not give ctx lines an sgr prefix, it seems more reliable to detect from the hunk.

  for line in lines:
    match = line.match
    kind = line.kind
    is_add_rem = (kind in ('rem', 'add'))
    if not is_prev_add_rem and is_add_rem: chunk_idx += 1
    is_prev_add_rem = is_add_rem
    if kind in ('ctx', 'rem', 'add'):
      line.is_src = True
      if kind == 'ctx':
        line.set_text(key='ctx_text', clip=is_loc_colored) # Clip is a hack; ctx lines do not have a leading color sequence.
      elif kind == 'rem':
        line.set_text(key='rem_text')
        line.chunk_idx = chunk_idx
        insert_unique_line(old_uniques, line.text, old_num)
      elif kind == 'add':
        line.set_text(key='add_text')
        line.chunk_idx = chunk_idx
        insert_unique_line(new_uniques, line.text, new_num)
      if kind in ('ctx', 'rem'):
        assert old_num not in old_lines
        assert old_num not in old_ctx_nums
        line.old_num = old_num
        old_lines[old_num] = line
        old_ctx_nums.add(old_num)
        old_num += 1
      if kind in ('ctx', 'add'):
        assert new_num not in new_lines
        assert new_num not in new_ctx_nums
        line.new_num = new_num
        new_lines[new_num] = line
        new_ctx_nums.add(new_num)
        new_num += 1
    elif kind == 'loc':
      is_loc_colored = bool(line.match['git_color'])
      o = int(match['old_num'])
      if o > 0:
        assert o > old_num, (o, old_num, match.string)
        old_num = o
      n = int(match['new_num'])
      if n > 0:
        assert n > new_num
        new_num = n
    elif kind == 'diff': # Not the best way to parse paths, because paths with spaces are ambiguous.
      paths = clip_reset(match['diff_paths']).split(' ') # Split into words, then guess at old and new split as best we can.
      i = len(paths) // 2 # Assume that both sides have the same number of spaces between them.
      # Note: if this does not prove sufficient for file renames we could try to find a split that matches either head or tail.
      old_path = ' '.join(paths[:i])
      new_path = ' '.join(paths[i:])
    elif kind == 'old': old_path = vscode_path(clip_reset(match['old_path']).rstrip('\t'))
    elif kind == 'new': new_path = vscode_path(clip_reset(match['new_path']).rstrip('\t')) # Not sure why this trailing tab appears.
    #^ These lines are a better way to parse the paths, but are not always present (particularly when one side is /dev/null).
    #^ Since they come after the diff line, they will overwrite the previous guess.

  # Detect moved lines.

  def diff_lines_match(old_idx:int, new_idx:int) -> bool:
    if old_idx in old_ctx_nums or new_idx in new_ctx_nums: return False
    try: return old_lines[old_idx].text.strip() == new_lines[new_idx].text.strip()
    except KeyError: return False

  old_moved_nums:Set[int] = set()
  new_moved_nums:Set[int] = set()
  for body, new_idx in new_uniques.items():
    if new_idx is None: continue
    old_idx = old_uniques.get(body)
    if old_idx is None: continue
    p_o = old_idx
    p_n = new_idx
    while diff_lines_match(p_o-1, p_n-1):
      p_o -= 1
      p_n -= 1
    e_o = old_idx + 1
    e_n = new_idx + 1
    while diff_lines_match(e_o, e_n):
      e_o += 1
      e_n += 1
    old_moved_nums.update(range(p_o, e_o))
    new_moved_nums.update(range(p_n, e_n))

  # Break lines into rem/add chunks and print them.
  # While a "hunk" is a series of (possibly many) ctx/rem/add lines provided by git diff,
  # a "chunk" is either a contiguous block of rem/add lines, or else any other single line.
  # This approach simplifies the token diffing process so that it is a reasonably
  # straightforward comparison of a rem block to an add block.

  def chunk_key(line:DiffLine) -> Tuple[bool, int, bool]:
    return (line.is_src, line.chunk_idx, (line.old_num in old_moved_nums or line.new_num in new_moved_nums))

  for ((is_src, chunk_idx, is_moved), _chunk) in groupby(lines, key=chunk_key):
    chunk = list(_chunk) # We iterate over the sequence several times.
    if chunk_idx and not is_moved: # Chunk should be diffed by tokens.
      # We must ensure that the same number of lines is output, at least for `-interactive` mode.
      # Currently, we do not reorder lines at all, but that is an option for the future.
      rem_lines = [l for l in chunk if l.old_num]
      add_lines = [l for l in chunk if l.new_num]
      add_token_diffs(rem_lines, add_lines)
    elif is_src: # ctx or moved.
      for l in chunk:
        l.text = ''.join(sanitize_for_console(l.text))

    # Print lines.
    for line in chunk:
      kind = line.kind
      match = line.match
      text = line.text
      if kind == 'ctx':
        print(text)
      elif kind == 'rem':
        m = C_REM_MOVED if line.old_num in old_moved_nums else ''
        print(m, text, C_END, sep='')
      elif kind == 'add':
        m = C_ADD_MOVED if line.new_num in new_moved_nums else ''
        print(m, text, C_END, sep='')
      elif kind == 'loc':
        new_num = match['new_num']
        snippet = clip_reset(match['parent_snippet'])
        s = ' ' + C_SNIPPET if snippet else ''
        print(C_LOC, new_path, ':', new_num, ':', s, snippet, C_END, sep='')
      elif kind == 'diff':
        msg = new_path if (old_path == new_path) else '{} -> {}'.format(old_path, new_path)
        print(C_FILE, msg, ':', C_END, sep='')
      elif kind == 'meta':
        print(C_MODE, new_path, ':', RST, ' ', line.text, sep='')
      elif kind in dropped_kinds:
        if interactive: # Cannot drop lines, because interactive mode slices the diff by line counts.
          print(C_DROPPED, line.text, RST, sep='')
      elif kind in pass_kinds:
        print(line.text)
      else:
        raise Exception('unhandled kind: {}\n{!r}'.format(kind, text))


dropped_kinds = {
  'idx', 'old', 'new'
}

pass_kinds = {
  'author', 'commit', 'date', 'empty', 'other'
}


def insert_unique_line(d:Dict[str, Optional[int]], line:str, idx:int) -> None:
  'For the purpose of movement detection, lines are tested for uniqueness after stripping leading and trailing whitespace.'
  body = line.strip()
  if body in d: d[body] = None
  else: d[body] = idx


def add_token_diffs(rem_lines:List[DiffLine], add_lines:List[DiffLine]) -> None:
  'Rewrite DiffLine.text values to include per-token diff highlighting.'
  r = HighlightState(lines=rem_lines, tokens=tokenize_difflines(rem_lines), hl_ctx=C_REM_CTX, hl_space=C_REM_SPACE, hl_token=C_REM_TOKEN)
  a = HighlightState(lines=add_lines, tokens=tokenize_difflines(add_lines), hl_ctx=C_ADD_CTX, hl_space=C_ADD_SPACE, hl_token=C_ADD_TOKEN)
  for r_r, r_a in calc_diff(r.tokens, a.tokens):
    if r_r and r_a: # Matching tokens; highlight as context.
      r.highlight_frags(r_r, is_ctx=True)
      a.highlight_frags(r_a, is_ctx=True)
    elif r_r: r.highlight_frags(r_r, is_ctx=False)
    elif r_a: a.highlight_frags(r_a, is_ctx=False)
  # Update the mutable lines lists.
  r.update_lines()
  a.update_lines()


H_START, H_CTX, H_SPACE, H_TOKEN = range(4)

class HighlightState:

  def __init__(self, lines:List[DiffLine], tokens:List[str], hl_ctx:str, hl_space:str, hl_token:str) -> None:
    self.lines = lines
    self.tokens = tokens
    self.hl_ctx = hl_ctx # Context highlight.
    self.hl_space = hl_space # Significant space highlight.
    self.hl_token = hl_token # Token highlighter.
    self.state = H_START
    self.line_idx = 0
    self.frags:List[List[str]] = [[] for _ in lines]

  def highlight_frags(self, rng:range, is_ctx:bool) -> None:
    for frag in self.tokens[rng.start:rng.stop]:
      line_frags = self.frags[self.line_idx]
      if frag == '\n':
        if self.state != H_CTX:
          line_frags.append(self.hl_ctx) # When combined with C_END, this highlights to end of line.
        self.state = H_START
        self.line_idx += 1
      else:
        if is_ctx:
          if self.state != H_CTX:
            self.state = H_CTX
            line_frags.append(self.hl_ctx)
        elif frag.isspace():
          if self.state == H_START: # Don't highlight spaces at the start of lines.
            self.state = H_TOKEN
            line_frags.append(self.hl_token)
          elif self.state == H_CTX:
            self.state = H_SPACE
            line_frags.append(self.hl_space)
        else:
          if self.state != H_TOKEN:
            self.state = H_TOKEN
            line_frags.append(self.hl_token)
        line_frags.extend(sanitize_for_console(frag))

  def update_lines(self) -> None:
    for line, line_frags in zip(self.lines, self.frags):
      line.text = ''.join(line_frags)


def tokenize_difflines(lines:List[DiffLine]) -> List[str]:
  'Convert the list of line texts into a single list of tokens, including newline tokens.'
  tokens:List[str] = []
  for line in lines:
    tokens.extend(m[0] for m in token_pat.finditer(line.text))
    tokens.append('\n')
  return tokens


def is_token_junk(token:str) -> bool:
  '''
  Treate newlines as tokens, but all other whitespace as junk.
  This forces the diff algorithm to respect line breaks but not get distracted aligning to whitespace.
  '''
  return token.isspace() and token != '\n'


git_diff_graph_mode_pat = re.compile(r'(?x) [ /\*\|\\]*') # space is treated as literal inside of brackets, even in extended mode.

diff_pat = re.compile(r'''(?x)
(?P<git_color> \x1b \[ \d* m)*
(?:
  (?P<empty>    $ )
| (?P<commit>   commit\ [0-9a-z]{40} )
| (?P<author>   Author: )
| (?P<date>     Date:   )
| (?P<diff>     diff\ --git\ (?P<diff_paths>.+) )
| (?P<idx>      index   )
| (?P<old>      ---     \ (?P<old_path>.+) )
| (?P<new>      \+\+\+  \ (?P<new_path>.+) )
| (?P<loc>      @@\ -(?P<old_num>\d+)(?P<old_len>,\d+)?\ \+(?P<new_num>\d+)(?P<new_len>,\d+)?\ @@
    (?:\x1b\[m)? \ ? (?:\x1b\[m)? (?P<parent_snippet>.*) ) # Note the RST SPACE RST sequence.
| (?P<ctx>      \  (?P<ctx_text>.*) )
| (?P<rem>      -  (?P<rem_text>.*) )
| (?P<add>      \+(?:\x1b\[m\x1b\[32m)? (?P<add_text>.*) ) # Hack to remove extra color sequences that git 2.19.2 shows for these lines only.
| (?P<meta>
  ( old\ mode
  | new\ mode
  | deleted\ file\ mode
  | new\ file\ mode
  | copy\ from
  | copy\ to
  | rename\ from
  | rename\ to
  | similarity\ index
  | dissimilarity\ index ) )
| (?P<other> .* )
)
''')


token_pat = re.compile(r'''(?x)
  (?:(?!_)\w)+ # Word characters, excluding underscores.
| \d+ # Numbers.
| \ + # Spaces; distinct from other whitespace.
| \t+ # Tabs; distinct from other whitespace.
| \s+ # Other whitespace.
| . # Any other single character; newlines are never present so DOTALL is irrelevant.
''')


# same-same colors.

C_FILE = sgr(BG, rgb6(1, 0, 1))
C_MODE = sgr(BG, rgb6(1, 0, 1))
C_LOC = sgr(BG, rgb6(0, 1, 2))
C_UNKNOWN = sgr(BG, rgb6(5, 0, 5))
C_SNIPPET = sgr(TXT, gray26(22))
C_DROPPED = sgr(TXT, gray26(10))

REM_BG = rgb6(1, 0, 0)
ADD_BG = rgb6(0, 1, 0)

C_REM_MOVED = sgr(BG, REM_BG, TXT, rgb6(4, 2, 0)) # Move detected.
C_ADD_MOVED = sgr(BG, ADD_BG, TXT, rgb6(2, 4, 0))

# Token highlighting.
C_REM_CTX = sgr(BG, REM_BG, RST_TXT, RST_BOLD)
C_ADD_CTX = sgr(BG, ADD_BG, RST_TXT, RST_BOLD)
C_REM_SPACE = sgr(BG, rgb6(3, 0, 0), RST_TXT, BOLD) # Change to space.
C_ADD_SPACE = sgr(BG, rgb6(0, 3, 0), RST_TXT, BOLD)
C_REM_TOKEN = sgr(BG, REM_BG, TXT, rgb6(5, 2, 3), BOLD)
C_ADD_TOKEN = sgr(BG, ADD_BG, TXT, rgb6(2, 5, 3), BOLD)

C_RST_TOKEN = sgr(RST_TXT, RST_BOLD)

C_END = FILL


def vscode_path(path:str) -> str:
  'VSCode will only recognize source locations if the path contains a slash; add "./" to plain file names.'
  if '/' in path or '<' in path or '>' in path: return path # Do not alter pseudo-names like <stdin>.
  return './' + path


def clip_reset(text:str) -> str:
  return text[:-len(reset_sgr)] if text.endswith(reset_sgr) else text

reset_sgr = '\x1b[m' # Git uses the short version with "0" code omitted.



def errL(*items:Any) -> None: print(*items, sep='', file=stderr)

def errSL(*items:Any) -> None: print(*items, file=stderr)


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import re
from argparse import ArgumentParser, FileType, Namespace
from os import environ
from shutil import copyfile
from typing import DefaultDict, NoReturn, Set

from ..diff import calc_diff
from ..fs import is_file, path_exists
from ..path import norm_path, path_dir, path_join, path_rel_to_dir


pat_version = '0'


def main() -> None:
  parser = ArgumentParser(prog='pat', description='create or apply a .pat patch file.')
  parser.epilog = "for help with a specific command, pass '-h' to that command."

  subs = parser.add_subparsers()
  subs.required = True
  subs.dest = 'command'


  sub_create = subs.add_parser('create',
    help='[original] [modified] [patch]: create an empty .pat file at [patch] referencing [original], and copy [original] to [modified].')

  sub_create.set_defaults(handler=main_create)

  sub_create.add_argument('original',
    help='source file to use as the original (left/minus) side of the patch.')

  sub_create.add_argument('modified',
    help='path at which to copy the modified (right/plus) side of the patch; must not exist.')

  sub_create.add_argument('patch',
    help='path at which to create the new empty .pat file; must not exist.')


  sub_diff = subs.add_parser('diff',
    help='[original] [modified] [out]: create .pat style diff between [original] and [modified], writing it to [out].')

  sub_diff.set_defaults(handler=main_diff)

  sub_diff.add_argument('original', type=FileType('r'),
    help='source file to use as the original (left/minus) side of the patch.')

  sub_diff.add_argument('modified', type=FileType('r'),
    help='source file to use as the modified (right/plus) side of the patch.')

  sub_diff.add_argument('out', type=FileType('w'),
    help='output path.')

  sub_diff.add_argument('-min-context', type=int, default=3,
    help='minimum number of lines of context to show before each hunk.')


  sub_apply = subs.add_parser('apply',
    help='[patch] [out]?: apply a [patch] to the source file specified in that patch, and write it to [out] or stdout.')

  sub_apply.set_defaults(handler=main_apply)

  sub_apply.add_argument('patch', type=FileType('r'),
    help='input .pat path to apply')

  sub_apply.add_argument('out', nargs='?', type=FileType('w'), default='-',
    help='output path (defaults to stdout)')

  args = parser.parse_args()
  args.handler(args)


def main_create(args: Namespace) -> None:
  'create command entry point.'
  original = args.original
  modified = args.modified
  patch = args.patch
  if not is_file(original, follow=True):  exit("pat create error: 'original' is not an existing file: " + original)
  if path_exists(modified, follow=False): exit("pat create error: 'modified' file already exists: " + modified)
  if path_exists(patch, follow=False):    exit("pat create error: 'patch' file already exists: " + patch)
  patch_dir = path_dir(patch)
  orig_rel = path_rel_to_dir(original, patch_dir)
  with open(patch, 'w') as f:
    f.write('pat v' + pat_version + '\n')
    f.write(orig_rel + '\n')
  copyfile(original, modified)


def main_diff(args) -> None:
  'diff command entry point.'
  original = args.original
  modified = args.modified
  f_out = args.out
  min_context = args.min_context

  out_dir = path_dir(f_out.name)
  orig_rel = path_rel_to_dir(original.name, out_dir)

  if min_context < 1: exit('min-context value must be positive.')

  if original.name.find('..') != -1:
    exit(f"original path cannot contain '..': {original.name!r}")

  o_lines = original.readlines()
  m_lines = modified.readlines()

  if o_lines and not o_lines[-1].endswith('\n'):
    exit(f'{original.name}:{len(o_lines)} original document is missing final newline (not yet supported).')
  if m_lines and not m_lines[-1].endswith('\n'):
    exit(f'{modified.name}:{len(m_lines)} modified document is missing final newline (not yet supported).')

  write = f_out.write

  write('pat v' + pat_version + '\n')
  write(orig_rel + '\n')

  line_indices = DefaultDict[str,Set[int]](set) # maps line contents to line numbers.
  for i, line in enumerate(o_lines):
    line_indices[line].add(i)

  diff = calc_diff(o_lines, m_lines) # List of (orig, mod) pairs of ranges.

  match_o = range(0, 0) # Previous match range on original side.
  em_end_o = 0 # End index of emitted original lines, either as 'match' or 'rem' chunks.

  has_start_symbol = False
  for r_o, r_m in diff:
    if r_o and r_m: # match range.
      match_o = r_o
      continue
    if not (r_o or r_m): # no-op.
      continue
    # Calculate how much context we need for this hunk to be unambiguous.
    # Work backwards from the end of either the current removed range or the previous match range.
    end_o = (r_o or match_o).stop
    last_o = end_o - 1
    matching_indices = line_indices[o_lines[last_o]] if (last_o >= 0) else set()
    # Iterate back through the first line of context.
    i = em_end_o
    for i in reversed(range(em_end_o, last_o)):
      decr_indices = [j-1 for j in matching_indices] # step all candidates backwards.
      curr_indices = line_indices[o_lines[i]]
      matching_indices = curr_indices.intersection(decr_indices)
      if len(matching_indices) == 1:
        break
    ctx_start_o = max(em_end_o, min(i, end_o - min_context))
    #errFL('* ci:{}', ci)
    if ctx_start_o == 0 and not has_start_symbol:
      has_start_symbol = True
      write('\n|^\n') # Add first separator line and start-of-file symbol line.
    elif em_end_o < ctx_start_o: # Not merged with previous hunk; separate with blank line.
      write('\n')
    # Output context and diff.
    for o in range(ctx_start_o, match_o.stop):
      write('| ' + o_lines[o]) # write context lines.
    for o in r_o:
      write('- ' + o_lines[o]) # remove lines from original.
    for m in r_m:
      write('+ ' + m_lines[m]) # add lines from modified.
    em_end_o = end_o


# version pattern is applied to the first line of documents;
# programs processing input strings may or may not check for a version as appropriate.
version_re = re.compile(r'pat v(\d+)\n')
def main_apply(args) -> None:
  'apply command entry point.'
  f_patch = args.patch
  f_out = args.out

  def patch_fail(line_num, msg) -> NoReturn:
    exit(f'{f_patch.name}:{line_num+1}: {msg}')

  version_line = f_patch.readline()
  orig_path_line = f_patch.readline()
  patch_lines = f_patch.readlines()

  m = version_re.fullmatch(version_line)
  if not m:
    patch_fail(0, f'first line should specify pat version matching pattern: {version_re.pattern!r}\n  found: {version_line!r}')
  version = m.group(1)
  if version != pat_version: patch_fail(0, f'unsupported version number: {version}')

  if not orig_path_line:
    patch_fail(1, 'patch file does not specify an original path')
  orig = orig_path_line.rstrip()

  if orig.startswith('/'):
    orig_path = norm_path(environ.get('PROJECT_DIR', '.')) + orig
  else:
    orig_path = norm_path(path_join(path_dir(f_patch.name), orig))
  try: f_orig = open(orig_path)
  except FileNotFoundError: patch_fail(1, f'could not open source path specified by patch: {orig_path!r}')

  orig_lines = f_orig.readlines()

  orig_line_indices = DefaultDict[str, Set[int]](set)
  for i, line in enumerate(orig_lines):
    orig_line_indices[line].add(i)

  len_orig = len(orig_lines)
  orig_index = 0
  def orig_line() -> str: return orig_lines[orig_index]

  for pi, patch_line in enumerate(patch_lines, 2): # advance two lines.
    if patch_line == '\n':
      continue
    if patch_line == '|^\n':
      if orig_index != 0:
        patch_fail(pi, 'patch start-of-file symbol `|^` may only occur at beginning of patch.')
      continue
    prefix = patch_line[0]
    line = patch_line[2:]
    if prefix == '#':
      pass
    elif prefix in '|-':
      orig_index_start = orig_index
      while orig_index < len_orig and orig_line() != line:
        f_out.write(orig_line())
        orig_index += 1
      if orig_index == len_orig:
        patch_fail(pi,
          f'patch context line does not match in source line range: {orig_index_start+1}-{orig_index+1}\n{patch_line}')
      if prefix == '|':
        f_out.write(orig_line())
      orig_index += 1
    elif prefix == '+':
      f_out.write(line)
    else:
      patch_fail(pi, f'bad patch line prefix: {prefix!r}')

  while orig_index < len_orig:
    f_out.write(orig_line())
    orig_index += 1


if __name__ == '__main__': main()
